"""
A module for working with averages.
"""

# built-in
from typing import List as _List


class MovingAverage:  # pylint: disable=too-many-instance-attributes
    """A class for managing a moving average of floats."""

    def __init__(self, depth: int = 10, initial: float = 0.0) -> None:
        """Initialize this moving average."""

        self._index: int = 0
        self._data: _List[float] = []
        self._sum: float = initial

        # Intentionally public members.
        self.depth: int = depth
        self.value: float = initial
        self.max: float = initial
        self.min: float = initial

        self.reset(initial=initial)
        self._initialized = False

    def __call__(self, value: float) -> float:
        """Add a new value to the dataset and get the average."""

        # Remove the oldest value.
        self._sum -= self._data[self._index]

        # Add the new value.
        self._sum += value

        # Calculate the new average.
        self.value = self._sum / self.depth

        # Ensure that max and min don't retain an initial value.
        if not self._initialized:
            self.max = value
            self.min = value
            self._initialized = True
        else:
            self.max = max(self.max, value)
            self.min = min(self.min, value)

        # Leave the new value behind.
        self._data[self._index] = value
        self._index += 1
        self._index %= self.depth

        return self.value

    def reset(self, initial: float = 0.0) -> None:
        """Reset the average value."""

        self._data = [initial for _ in range(self.depth)]
        self._sum = sum(self._data)
        self.value = self._sum / self.depth
        self.max = initial
        self.min = initial
        self._initialized = False

    def resize(self, depth: int, initial: float = 0.0) -> None:
        """Set a new depth for this moving average and reset the value."""

        self.depth = depth
        self.reset(initial=initial)



"""
A module implementing various data-file encoders.
"""

# built-in
from configparser import ConfigParser
from json import dump
from logging import Logger, getLogger
from typing import cast as _cast

# third-party
from tomli_w import dumps

# internal
from vcorelib import DEFAULT_INDENT as _DEFAULT_INDENT
from vcorelib.dict import GenericStrDict as _GenericStrDict
from vcorelib.dict import consume
from vcorelib.io.types import DataStream as _DataStream
from vcorelib.io.types import JsonObject as _JsonObject
from vcorelib.io.types import YAML_INTERFACE as _YAML_INTERFACE
from vcorelib.math.time import TIMER as _TIMER

_LOG = getLogger(__name__)


def encode_json(
    configs: _JsonObject, ostream: _DataStream, _: Logger = _LOG, **kwargs
) -> int:
    """Write config data as JSON to the output stream."""

    with _TIMER.measure_ns() as token:
        # Normalize arguments with some defaults.
        dump(
            configs,
            ostream,
            indent=consume(kwargs, "indent", _DEFAULT_INDENT),
            sort_keys=consume(kwargs, "sort_keys", True),
            **kwargs,
        )
    return _TIMER.result(token)


def encode_yaml(
    configs: _JsonObject, ostream: _DataStream, _: Logger = _LOG, **kwargs
) -> int:
    """Write config data as YAML to the output stream."""

    with _TIMER.measure_ns() as token:
        _YAML_INTERFACE.dump(configs, ostream, **kwargs)
    return _TIMER.result(token)


def encode_ini(
    configs: _JsonObject, ostream: _DataStream, _: Logger = _LOG, **kwargs
) -> int:
    """Write config data as INI to the output stream."""

    with _TIMER.measure_ns() as token:
        cparser = ConfigParser(
            interpolation=consume(kwargs, "interpolation"), **kwargs
        )
        cparser.read_dict(_cast(_GenericStrDict, configs))
        cparser.write(ostream)
    return _TIMER.result(token)


def encode_toml(
    configs: _JsonObject, ostream: _DataStream, _: Logger = _LOG, **kwargs
) -> int:
    """Write config data as TOML to the output stream."""

    with _TIMER.measure_ns() as token:
        ostream.write(dumps(configs, **kwargs))
    return _TIMER.result(token)



"""
Math utilities involving units of time and other conversions.
"""

# built-in
from contextlib import contextmanager
from io import StringIO
from logging import INFO as _INFO
from logging import Logger as _Logger
from logging import LoggerAdapter as _LoggerAdapter
from math import floor as _floor
from time import perf_counter_ns as _perf_counter_ns
from time import time_ns as _time_ns
from typing import Dict as _Dict
from typing import Iterator as _Iterator
from typing import Tuple as _Tuple
from typing import Union as _Union

# internal
from vcorelib.math import KIBI_UNITS as _KIBI_UNITS
from vcorelib.math import SI_UNITS as _SI_UNITS
from vcorelib.math import UnitSystem as _UnitSystem
from vcorelib.math import unit_traverse as _unit_traverse


def default_time_ns() -> int:
    """Get a timestamp value using a default method."""
    return _time_ns()


def seconds_str(seconds: int) -> _Tuple[str, int]:
    """
    Attempt to characterize a large amount of seconds into a string describing
    hours and minutes, returning the string (may be empty) and the seconds
    left over.
    """

    result = ""

    minutes = 0
    if seconds >= 60:
        minutes = seconds // 60
        seconds = seconds % 60
        result = f"{minutes}m"

    if minutes >= 60:
        hours = minutes // 60
        minutes = minutes % 60
        result = f"{hours}h {minutes}m"

    return result, seconds


def nano_str(
    nanos: int,
    is_time: bool = False,
    max_prefix: int = 3,
    unit: _UnitSystem = _SI_UNITS,
    prefix_space: bool = False,
    iteration: int = 0,
) -> str:
    """
    Convert an arbitrary value in a 10^-9 domain into as concise of a string
    as possible.
    """

    decimal, fractional, prefix = _unit_traverse(
        nanos, unit, max_prefix, iteration
    )

    with StringIO() as stream:
        if not prefix and is_time:
            result, decimal = seconds_str(decimal)
            stream.write(result)
            if result:
                stream.write(" ")

        # Normalize the fractional component if necessary.
        if unit.divisor != 1000 and fractional != 0:
            fractional = _floor(float(fractional / unit.divisor) * 1000.0)

        stream.write(str(decimal))
        if fractional:
            stream.write(f".{fractional:03}")
        if prefix_space:
            stream.write(" ")
        stream.write(prefix)
        return stream.getvalue()


def byte_count_str(byte_count: int) -> str:
    """Get a string representing a number of bytes."""
    return nano_str(byte_count, False, 99, _KIBI_UNITS, True) + "B"


LoggerType = _Union[_Logger, _LoggerAdapter]


class Timer:
    """A class for measuring and logging how long events take."""

    def __init__(self) -> None:
        """Initialize this timer."""

        self.curr: int = 0
        self.data: _Dict[int, int] = {}

    @contextmanager
    def measure_ns(self) -> _Iterator[int]:
        """
        Compute the time that the caller's context takes, provides an integer
        token that can be used to query for the result afterwards.
        """

        curr = self.curr
        self.curr += 1
        start = _perf_counter_ns()
        try:
            yield curr
        finally:
            self.data[curr] = _perf_counter_ns() - start

    def result(self, token: int) -> int:
        """Get the timer result."""
        return self.data.pop(token, -1)

    @contextmanager
    def log(
        self,
        log: LoggerType,
        message: str,
        *args,
        level: int = _INFO,
        **kwargs,
    ) -> _Iterator[None]:
        """Log how long the caller's context took to execute."""

        with self.measure_ns() as token:
            yield
        result = self.result(token)

        # Log the duration spent yielded.
        log.log(
            level,
            message + " completed in %ss.",
            *args,
            nano_str(result, True),
            **kwargs,
        )


TIMER = Timer()



"""
A module for working with graph nodes.
"""

# built-in
from io import StringIO as _StringIO
from typing import TextIO as _TextIO
from typing import Type as _Type

# internal
from vcorelib.graph.abc import AbstractDiGraphNode as _AbstractDiGraphNode
from vcorelib.graph.abc import T as _T
from vcorelib.graph.abc import V as _V
from vcorelib.graph.edge import write_attributes as _write_attributes
from vcorelib.graph.edge import write_node_id as _write_node_id


class DiGraphNode(_AbstractDiGraphNode):
    """A base implementation for a directed-graph node."""

    def to_stream(self, stream: _TextIO, **kwargs) -> None:
        """Write this object to a text stream."""

        _write_node_id(self.label, stream, port=self.port)

        if "label" not in self.node_attributes:
            label_parts = [self.label]
            with _StringIO() as label_stream:
                self.ports.label(label_stream)
                label_parts.append(label_stream.getvalue())
            self.node_attributes["label"] = "|".join(label_parts)

        _write_attributes(stream, data=self.node_attributes)

        stream.write(";")

    def add_child(
        self,
        label: str,
        node: _T = None,
        graph_kind: _Type[_V] = None,
        **kwargs,
    ) -> _T:
        """
        Add an edge between this node (source) and the other node
        (destination).
        """

        graph = self.graph(kind=graph_kind)

        # Ensure the other node is a vertex in the graph.
        node = graph.handle_node(label, node=node)
        graph.add_edge(self.label, node.label, **kwargs)
        return node

    def add_parent(
        self,
        label: str,
        node: _T = None,
        graph_kind: _Type[_V] = None,
        **kwargs,
    ) -> _T:
        """
        Add an edge between this node (destination) and the other node
        (source).
        """

        graph = self.graph(kind=graph_kind)

        # Ensure the other node is a vertex in the graph.
        node = graph.handle_node(label, node=node)
        graph.add_edge(node.label, self.label, **kwargs)
        return node

    def add_parallel(
        self,
        label: str,
        node: _T = None,
        graph_kind: _Type[_V] = None,
        **kwargs,
    ) -> _T:
        """
        Add a parallel edge between two nodes. This instance is used as
        'node1'.
        """

        graph = self.graph(kind=graph_kind)

        # Ensure the other node is a vertex in the graph.
        node = graph.handle_node(label, node=node)
        graph.add_parallel(self.label, node.label, **kwargs)
        return node



"""
A module for implementing tasks in a dependency tree.
"""

from __future__ import annotations

# built-in
import asyncio as _asyncio
from contextlib import ExitStack as _ExitStack
from json import dumps as _dumps
from logging import Logger, getLogger
from os import linesep as _linesep
from typing import Any as _Any
from typing import Callable as _Callable
from typing import Coroutine as _Coroutine
from typing import Dict as _Dict
from typing import Iterable as _Iterable
from typing import List as _List
from typing import Optional as _Optional
from typing import Set as _Set
from typing import Tuple as _Tuple

# internal
from vcorelib.dict import merge_dicts
from vcorelib.math.time import Timer, nano_str
from vcorelib.target import Substitutions, Target

Outbox = _Dict[str, _Any]
Inbox = _Dict[str, Outbox]
TaskExecute = _Callable[[Inbox, Outbox], _Coroutine[_Any, _Any, bool]]
TaskGenerator = _Callable[..., _asyncio.Task]


class TaskFailed(Exception):
    """A custom exception to indicate that a task failed."""


class Task:  # pylint: disable=too-many-instance-attributes
    """A basic task interface definition."""

    stages = ["run_enter", "run", "run_exit"]
    default_requirements: _Set[str] = set()

    def __init__(
        self,
        name: str,
        *args,
        execute: TaskExecute = None,
        log: Logger = None,
        timer: Timer = None,
        target: Target = None,
        **kwargs,
    ) -> None:
        """Initialize this task."""

        self.name = name
        self.inbox: Inbox = {}
        self.outbox: Outbox = {}
        self.dependencies: _List[TaskGenerator] = []
        self._stack: _Optional[_ExitStack] = None

        # Dependency resolution state.
        self._resolved = False
        self.literals: _Set[str] = set()

        # Invocation state.
        self._continue = True
        self._running: _Set[str] = set()
        self._to_signal: int = 0
        self._sem: _Optional[_asyncio.Semaphore] = None

        # Metrics.
        self.times_invoked: int = 0
        self.dependency_time: int = 0
        self.execute_time: int = 0
        if timer is None:
            timer = Timer()
        self.timer = timer

        # Get a logger if needed.
        if log is None:
            log = getLogger(name)
        self.log = log

        # Create an execute function if none was provided.
        if execute is None:
            execute = self.create_execute(*args, **kwargs)
        self.execute = execute

        # Create a target if one wasn't provided.
        if target is None:
            target = Target(self.name)
        self.target: Target = target

    def __str__(self) -> str:
        """Convert this task into a string."""
        return self.name

    @property
    def stack(self) -> _ExitStack:
        """Get this task's execution stack."""
        assert self._stack is not None
        return self._stack

    def resolved(
        self, compiled: str, substitutions: Substitutions = None
    ) -> bool:
        """Override this in a derived task to implement more complex logic."""

        # Only a literal target has a boolean resolved state.
        if self.target.literal:
            return self._resolved

        # This task is only resolved if the compiled literal appears in the
        # completed set.
        assert substitutions is not None
        return compiled in self.literals

    def resolve(
        self, log: Logger, compiled: str, substitutions: Substitutions = None
    ) -> None:
        """Mark this task resolved."""

        self._running.remove(compiled)
        self.literals.add(compiled)

        # A literal target only needs to be resovled once.
        if self.target.literal:
            self._resolved = True

        # A non-literal task must have valid substitutions.
        else:
            assert substitutions is not None

        # Signal to any other active tasks that this is complete.
        assert self._sem is not None
        for _ in range(self._to_signal):
            self._sem.release()
        if self._to_signal:
            log.debug("Signaled %d waiting tasks.", self._to_signal)
        self._to_signal = 0

    async def run_enter(
        self, _inbox: Inbox, _outbox: Outbox, *_args, **_kwargs
    ) -> bool:
        """A default enter method."""
        assert self
        return True

    async def run_exit(
        self, _inbox: Inbox, _outbox: Outbox, *_args, **_kwargs
    ) -> bool:
        """A default exit method."""
        assert self
        return True

    async def run(self, inbox: Inbox, outbox: Outbox, *args, **kwargs) -> bool:
        """Override this method to implement the task."""

        self.log.info(
            "inbox=%s, outbox=%s, args=%s, kwargs=%s",
            inbox,
            outbox,
            args,
            kwargs,
        )
        return True

    def create_execute(self, *args, **kwargs) -> TaskExecute:
        """Create a default execute function for this task."""

        async def wrapper(
            inbox: Inbox, outbox: Outbox, **substitutions
        ) -> bool:
            """A default implementation for a basic task. Override this."""

            merged = merge_dicts([{}, kwargs, substitutions], logger=self.log)

            # Run through all of the stages.
            result = True
            self._continue = True
            for stage in self.stages:
                if result and self._continue:
                    result = await getattr(self, stage)(
                        inbox, outbox, *args, **merged
                    )
            return result

        return wrapper

    def depend_on(
        self, task: "Task", eloop: _asyncio.AbstractEventLoop = None, **kwargs
    ) -> bool:
        """
        Register other tasks' output data to your input box. Return true
        if a new dependency was added.
        """

        def task_factory(**substitutions) -> _asyncio.Task[_Any]:
            """
            Create a task while injecting additional keyword arguments.
            """

            nonlocal eloop
            if eloop is None:
                eloop = _asyncio.get_running_loop()

            return eloop.create_task(
                task.dispatch(
                    self,
                    substitutions={**substitutions},
                    **kwargs,
                )
            )

        self.inbox[task.name] = task.outbox
        self.dependencies.append(task_factory)

        return True

    def depend_on_all(self, tasks: _Iterable["Task"], **kwargs) -> int:
        """
        Register multiple dependencies. Return the number of dependencies
        added.
        """

        added_count = 0
        for task in tasks:
            added_count += int(self.depend_on(task, **kwargs))
        return added_count

    async def resolve_dependencies(self, **substitutions) -> None:
        """
        A default dependency resolver for tasks. Note that the default
        dependency resolver cannot propagate current-task substitutions to
        its dependencies as they've already been explicitly registered.
        """
        await _asyncio.gather(*[x(**substitutions) for x in self.dependencies])

    def task_fail(
        self,
        compiled: str,
        substitutions: Substitutions,
        caller: "Task" = None,
        indent: int = 4,
    ) -> TaskFailed:
        """Build an exception message for when this task fails."""

        lines = [
            f"Task '{compiled}' failed in {nano_str(self.execute_time)}s."
        ]
        if caller is not None:
            lines.append(f"Called by: '{caller}'")
        lines.append(_dumps(substitutions, indent=indent))
        return TaskFailed(_linesep.join(lines))

    def dispatch_init(
        self,
        caller: "Task" = None,
        substitutions: Substitutions = None,
        **kwargs,
    ) -> _Tuple[Substitutions, str]:
        """Perform some dispatch initialization."""

        self.times_invoked += 1

        if caller is not None:
            self.log.debug("triggered by '%s'", caller)

        if substitutions is None:
            substitutions = {}

        # Merge substitutions in with other command-line arguments.
        merged = merge_dicts([{}, substitutions, kwargs], logger=self.log)
        compiled = self.target.compile(merged)

        # Wire our outbox to the caller's inbox.
        if caller is not None:
            caller.inbox[compiled] = self.outbox

        return merged, compiled

    async def dispatch(
        self,
        caller: "Task" = None,
        init_only: bool = False,
        substitutions: Substitutions = None,
        **kwargs,
    ) -> None:
        """Dispatch this task and return whether or not it succeeded."""

        merged, compiled = self.dispatch_init(
            caller=caller, substitutions=substitutions, **kwargs
        )

        # Return early if this task has already been executed.
        if self.resolved(compiled, merged):
            return

        if self._sem is None:
            self._sem = _asyncio.Semaphore(0)

        log = getLogger(compiled)
        if merged and self.times_invoked == 1:
            log.debug("substitutions: '%s'", merged)

        try:
            # If this task is already running, wait for it to complete.
            if compiled in self._running:
                self._to_signal += 1
                log.debug("Waiting for existing '%s' to resolve.", compiled)
                await self._sem.acquire()
                return

            # Signal to other task invocations that this instance is running.
            self._running.add(compiled)

            # Reset the outbox for this execution. Don't re-assign it or it
            # will lose its association with anything depending on it.
            for key in list(self.outbox.keys()):
                self.outbox.pop(key)

            # Wait for dependencies to finish processing.
            with self.timer.measure_ns() as token:
                await self.resolve_dependencies(**merged)
            self.dependency_time = self.timer.result(token)
            log.debug("dependencies: %ss", nano_str(self.dependency_time))

            # Allow a dry-run pass get only this far (before executing).
            if init_only:
                return

            with self.timer.measure_ns() as token:
                # Execute this task and don't propagate to tasks if this task
                # failed.
                with _ExitStack() as stack:
                    self._stack = stack
                    result = await self.execute(
                        self.inbox, self.outbox, **merged
                    )
            self.execute_time = self.timer.result(token)

            # Raise an exception if this task failed.
            if not result:
                raise self.task_fail(compiled, merged, caller)

            log.debug("execute: %ss", nano_str(self.execute_time))

            # Track that this task was resolved.
            self.resolve(log, compiled, merged)

        # Handle cancellation to simplify cleaning up the event loop.
        except _asyncio.CancelledError:
            log.warning("Task '%s' was cancelled!", self.name)


class Phony(Task):
    """
    A task stub that doesn't do anything. Useful for top-level targets that
    depend on other concrete tasks.
    """

    async def run(self, inbox: Inbox, outbox: Outbox, *args, **kwargs) -> bool:
        """Override this method to implement the task."""

        # Pass the inbox items through to the outbox. This makes phony tasks
        # much more useful.
        for key, value in inbox.items():
            outbox[key] = value
        return True


class FailTask(Task):
    """A task that always fails."""

    async def run(self, inbox: Inbox, outbox: Outbox, *args, **kwargs) -> bool:
        """Task fails by default."""
        assert self.stack
        return False



"""
A task definition for wrapping subprocess's 'run' method.
"""

# built-in
from asyncio import CancelledError as _CancelledError
from asyncio import create_subprocess_exec, create_subprocess_shell
from asyncio.subprocess import PIPE as _PIPE
from asyncio.subprocess import Process as _Process
from logging import Logger as _Logger
from platform import system as _system
import signal as _signal
from sys import executable as _executable
from typing import List as _List
from typing import Tuple as _Tuple

# internal
from vcorelib.paths import rel as _rel
from vcorelib.task import Inbox as _Inbox
from vcorelib.task import Outbox, Task


def is_windows() -> bool:
    """Determine if the current platform is Windows or not."""
    return _system() == "Windows"


def reconcile_platform(
    program: str, args: _List[str]
) -> _Tuple[str, _List[str]]:
    """
    Handle arguments for Windows. You cannot run a program directly on Windows
    under any circumstance, so pass arguments through to the shell.
    """

    args = ["/c", program] + args if is_windows() else args
    program = "cmd.exe" if is_windows() else program
    return program, args


async def handle_process_cancel(
    proc: _Process,
    name: str,
    logger: _Logger,
    stdin: bytes = None,
    signal: int = None,
) -> None:
    """
    Communicate with a process and send a signal to it if this task gets
    cancelled.
    """

    # Default to a valid interrupt signal.
    if signal is None:
        signal = getattr(_signal, "CTRL_C_EVENT", _signal.SIGINT)
    assert signal is not None

    try:
        await proc.communicate(input=stdin)

    except _CancelledError:
        # Send the process a signal and wait for it to terminate.
        proc.send_signal(signal)
        logger.warning("Sending signal %d to process '%s'.", signal, name)
        await proc.wait()
        raise

    finally:
        if proc.returncode is not None:
            logger.info("Process '%s' exited %d.", name, proc.returncode)


class SubprocessLogMixin(Task):
    """
    A class for creating asyncio subprocesses and logging what gets created.
    """

    async def subprocess_exec(
        self,
        program: str,
        *caller_args,
        args: str = "",
        separator: str = "::",
        stdout: int = None,
        stderr: int = None,
    ) -> _Process:
        """
        Create a process using subprocess exec and log what the arguments were.
        """

        exec_args = [x for x in args.split(separator) + [*caller_args] if x]

        # Use relative paths when logging to reduce output.
        self.log.info(
            "exec '%s': %s",
            str(_rel(program)),
            " ".join(str(_rel(x)) for x in exec_args),
        )

        program, exec_args = reconcile_platform(program, exec_args)
        proc = await create_subprocess_exec(
            program,
            *exec_args,
            stdout=stdout,
            stderr=stderr,
        )
        return proc

    async def exec(
        self,
        program: str,
        *caller_args,
        args: str = "",
        separator: str = "::",
        stdout: int = None,
        stderr: int = None,
    ) -> bool:
        """Execute a command and return whether or not it succeeded."""

        proc = await self.subprocess_exec(
            program,
            *caller_args,
            args=args,
            separator=separator,
            stdout=stdout,
            stderr=stderr,
        )
        await handle_process_cancel(proc, self.name, self.log)
        return proc.returncode == 0

    async def subprocess_shell(
        self,
        cmd: str,
        *caller_args,
        args: str = "",
        joiner: str = " ",
        separator: str = "::",
        stdout: int = None,
        stderr: int = None,
    ) -> _Process:
        """
        Create a process using subprocess shell and log what the command is.
        """

        command = cmd + joiner.join(
            x for x in args.split(separator) + [*caller_args] if x
        )
        self.log.info("shell: '%s'", command)
        proc = await create_subprocess_shell(
            command, stdout=stdout, stderr=stderr
        )
        return proc

    async def shell(
        self,
        cmd: str,
        *caller_args,
        args: str = "",
        joiner: str = " ",
        separator: str = "::",
        stdout: int = None,
        stderr: int = None,
    ) -> bool:
        """Execute a shell command and return whether or not it succeeded."""

        proc = await self.subprocess_shell(
            cmd,
            *caller_args,
            args=args,
            joiner=joiner,
            separator=separator,
            stdout=stdout,
            stderr=stderr,
        )
        await handle_process_cancel(proc, self.name, self.log)
        return proc.returncode == 0


class SubprocessExec(SubprocessLogMixin):
    """A task wrapping a subprocess."""

    async def run(
        self,
        inbox: _Inbox,
        outbox: Outbox,
        *caller_args,
        args: str = "--version",
        program: str = _executable,
        separator: str = "::",
        require_success: bool = True,
        **kwargs,
    ) -> bool:
        """
        Create a subprocess, wait for it to exit and add results to the outbox.
        """

        proc = await self.subprocess_exec(
            program,
            *caller_args,
            args=args,
            separator=separator,
            stdout=_PIPE,
            stderr=_PIPE,
        )
        await handle_process_cancel(proc, self.name, self.log)
        outbox["stdout"] = proc.stdout
        outbox["stderr"] = proc.stderr
        outbox["code"] = proc.returncode

        return True if not require_success else proc.returncode == 0


class SubprocessExecStreamed(SubprocessLogMixin):
    """A task wrapping a subprocess."""

    async def run(
        self,
        inbox: _Inbox,
        outbox: Outbox,
        *caller_args,
        args: str = "--version",
        program: str = _executable,
        separator: str = "::",
        require_success: bool = True,
        **kwargs,
    ) -> bool:
        """
        Create a subprocess, wait for it to exit and add results to the outbox.
        """

        proc = await self.subprocess_exec(
            program, *caller_args, args=args, separator=separator
        )
        await handle_process_cancel(proc, self.name, self.log)
        outbox["code"] = proc.returncode

        return True if not require_success else proc.returncode == 0


class SubprocessShell(SubprocessLogMixin):
    """A task wrapping a shell command."""

    async def run(
        self,
        inbox: _Inbox,
        outbox: Outbox,
        *caller_args,
        args: str = "",
        cmd: str = f"{_executable} --version",
        joiner: str = " ",
        separator: str = "::",
        require_success: bool = True,
        **kwargs,
    ) -> bool:
        """
        Run a shell command, wait for it to exit and add results to the outbox.
        """

        proc = await self.subprocess_shell(
            cmd,
            *caller_args,
            args=args,
            joiner=joiner,
            separator=separator,
            stdout=_PIPE,
            stderr=_PIPE,
        )
        await handle_process_cancel(proc, self.name, self.log)
        outbox["stdout"] = proc.stdout
        outbox["stderr"] = proc.stderr
        outbox["code"] = proc.returncode

        return True if not require_success else proc.returncode == 0


class SubprocessShellStreamed(SubprocessLogMixin):
    """A task wrapping a shell command."""

    async def run(
        self,
        inbox: _Inbox,
        outbox: Outbox,
        *caller_args,
        args: str = "",
        cmd: str = f"{_executable} --version",
        joiner: str = " ",
        separator: str = "::",
        require_success: bool = True,
        **kwargs,
    ) -> bool:
        """
        Run a shell command, wait for it to exit and add results to the outbox.
        """

        proc = await self.subprocess_shell(
            cmd, *caller_args, args=args, joiner=joiner, separator=separator
        )
        await handle_process_cancel(proc, self.name, self.log)
        outbox["code"] = proc.returncode

        return True if not require_success else proc.returncode == 0



"""
Implements management of target objects.
"""

# built-in
from typing import Any as _Any
from typing import Dict as _Dict
from typing import Iterable as _Iterable
from typing import Iterator as _Iterator
from typing import List as _List
from typing import NamedTuple as _NamedTuple
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _Union

# internal
from vcorelib.target import LITERAL_MATCH as _LITERAL_MATCH
from vcorelib.target import NO_MATCH as _NO_MATCH
from vcorelib.target import Target, TargetMatch


class TargetResolution(_NamedTuple):
    """A return type for the target resolver."""

    result: TargetMatch
    data: _Optional[_Any] = None

    def __bool__(self) -> bool:
        """Return whether or not this target resolution is a match."""
        return self.result.matched


NOT_RESOLVED = TargetResolution(_NO_MATCH, None)


class TargetResolver:
    """
    A class for registering target prototypes that can be used to match
    incoming data.
    """

    def __init__(self) -> None:
        """Initialize this target resolver."""

        self.literals: _Dict[str, _Any] = {}
        self.dynamic: _Dict[Target, _Any] = {}

    def register(self, data: str, value: _Any = None) -> bool:
        """
        Register a target to this resolver. If it is ever resolved in
        evaluation, value will be returned. Return whether or not any target
        was registered.
        """

        added = False
        target = Target(data)

        # Don't allow double registration but let the caller handle this.
        if target.literal:
            if data not in self.literals:
                self.literals[data] = value
                added = True
        elif target not in self.dynamic:
            self.dynamic[target] = value
            added = True

        return added

    def evaluate(self, data: str) -> TargetResolution:
        """Find the target that matches data, if one can be found."""

        # Optimize matching candidate data against many targets by first
        # testing the literal set.
        if data in self.literals:
            return TargetResolution(_LITERAL_MATCH, self.literals[data])

        matches: _List[_Tuple[Target, TargetMatch, _Any]] = []
        for candidate, value in self.dynamic.items():
            test = candidate.evaluate(data)
            if test.matched:
                matches.append((candidate, test, value))

        # If we find any dynamic match, ensure that we only matched a single
        # target.
        if matches:
            assert len(matches) == 1, (
                f"Matched '{data}' to {len(matches)} targets: "
                f"{', '.join(x[0].data for x in matches)}!"
            )
            return TargetResolution(matches[0][1], matches[0][2])

        return NOT_RESOLVED

    def evaluate_all(
        self, data: _Iterable[str]
    ) -> _Iterator[_Union[TargetResolution, str]]:
        """
        Evaluate all targets and optionally enforce that they all matched.
        """
        for item in data:
            evaluation = self.evaluate(item)
            yield evaluation if evaluation.result.matched else item



"""
Common type definitions for data encoding and decoding interfaces.
"""

# built-in
from enum import Enum
from io import StringIO
from logging import Logger
from pathlib import Path
from typing import Callable as _Callable
from typing import Dict as _Dict
from typing import Iterator as _Iterator
from typing import List as _List
from typing import NamedTuple
from typing import Optional as _Optional
from typing import TextIO
from typing import Tuple as _Tuple
from typing import Union as _Union

# third-party
from ruamel.yaml import YAML

# internal
from vcorelib.dict import merge
from vcorelib.paths import Pathlike as _Pathlike
from vcorelib.paths import get_file_ext

DEFAULT_ARCHIVE_EXT = "tar.gz"
DEFAULT_DATA_EXT = "json"

# A simple type system for JSON.
JsonPrimitive = _Union[str, int, float, bool, None]
JsonValue = _Union[
    JsonPrimitive, _Dict[str, JsonPrimitive], _List[JsonPrimitive]
]
JsonArray = _List[JsonValue]
JsonObject = _Dict[str, JsonValue]


class FileExtension(Enum):
    """A mapping of expected encoding type to file extensions."""

    UNKNOWN: _List[str] = ["unknown"]
    # Data formats.
    JSON: _List[str] = [DEFAULT_DATA_EXT]
    YAML: _List[str] = ["yaml", "yml", "eyaml"]
    INI: _List[str] = ["ini", "cfg"]
    TOML: _List[str] = ["toml"]
    # Archive formats.
    ZIP: _List[str] = ["zip"]
    TAR: _List[str] = [
        DEFAULT_ARCHIVE_EXT,
        "tar",
        "tar.bz2",
        "tar.lzma",
        "tar.xz",
    ]
    # Template formats.
    JINJA: _List[str] = ["j2", "jinja", "j2_template", "j2_macro"]

    def __str__(self) -> str:
        """Get this extension as a string."""
        return self.value[0]

    def is_template(self) -> bool:
        """Determine if this extension is a kind of template."""
        return self in {FileExtension.JINJA}

    def is_archive(self) -> bool:
        """Determine if this extension is a kind of archive file."""
        return self in {FileExtension.ZIP, FileExtension.TAR}

    @staticmethod
    def has_archive(path: Path) -> _Optional[Path]:
        """Determine if a path has an associated archive file."""

        for ext in [FileExtension.ZIP, FileExtension.TAR]:
            for ext_str in ext.value:  # pylint: disable=not-an-iterable
                check_path = Path(f"{path}.{ext_str}")
                if check_path.is_file():
                    return check_path

        return None

    def is_data(self) -> bool:
        """Determine if this etension is a kind of data file."""
        return self in {
            FileExtension.JSON,
            FileExtension.YAML,
            FileExtension.INI,
        }

    @staticmethod
    def from_ext(ext_str: str) -> _Optional["FileExtension"]:
        """Given a file extension, determine what kind of file it is."""

        result = FileExtension.UNKNOWN
        for ext in FileExtension:
            if ext_str in ext.value:
                result = ext
                break
        return None if result is FileExtension.UNKNOWN else result

    @staticmethod
    def from_path(
        path: _Pathlike, maxsplit: int = 1
    ) -> _Optional["FileExtension"]:
        """Get a known file extension for a path, if it exists."""
        return FileExtension.from_ext(get_file_ext(path, maxsplit=maxsplit))

    def apply(self, path: Path) -> Path:
        """Apply this extension's suffix to the given path."""
        return path.with_suffix("." + str(self))

    def candidates(
        self, path: Path, exists_only: bool = False
    ) -> _Iterator[Path]:
        """
        For a given path, iterate over candidate paths that have the suffixes
        for this kind of file extension.
        """
        for ext in self.value:
            path = path.with_suffix(f".{ext}")
            if not exists_only or path.exists():
                yield path

    @staticmethod
    def archive_candidates(
        path: Path, exists_only: bool = False
    ) -> _Iterator[Path]:
        """
        Iterate over all file extensions that could point to an archive file.
        """
        for file_ext in FileExtension:
            if file_ext.is_archive():
                for candidate in file_ext.candidates(path, exists_only):
                    yield candidate

    @staticmethod
    def data_candidates(
        path: Path, exists_only: bool = False
    ) -> _Iterator[Path]:
        """
        Iterate over all file extensions that could point to a data file.
        """
        for file_ext in FileExtension:
            if file_ext.is_data():
                for candidate in file_ext.candidates(path, exists_only):
                    yield candidate


class LoadResult(NamedTuple):
    """
    An encapsulation of the result of loading raw data, the data collected and
    whether or not it succeeded.
    """

    data: JsonObject
    success: bool
    time_ns: int = -1

    def __eq__(self, other: object) -> bool:
        """Don't compare timing when checking equivalence."""
        assert isinstance(other, (LoadResult, tuple))
        return bool(self.data == other[0] and self.success == other[1])

    def require_success(self, path: _Union[Path, str]) -> None:
        """Raise a canonical exception if this result is a failure."""
        assert self.success, f"Couldn't load '{path}'!"

    def merge(self, other: "LoadResult") -> "LoadResult":
        """Merge two load results."""

        # Add the time fields up if they're both positive.
        time_ns = self.time_ns
        if time_ns > 0 and other.time_ns > 0:
            time_ns += other.time_ns

        return LoadResult(
            merge(self.data, other.data),
            self.success and other.success,
            time_ns,
        )


EncodeResult = _Tuple[bool, int]
DataStream = _Union[TextIO, StringIO]
StreamProcessor = _Callable[[DataStream], DataStream]
DataDecoder = _Callable[[DataStream, Logger], LoadResult]
DataEncoder = _Callable[[JsonObject, DataStream, Logger], int]

# Only create the interface one so it's not re-created on every read and write
# attempt.
YAML_INTERFACE = YAML(typ="safe")



"""
A module for interacting with JSON schemas.
"""

# built-in
from logging import getLogger as _getLogger
from typing import Any as _Any
from typing import Type as _Type
from urllib.parse import urlparse as _urlparse

# third-party
from fastjsonschema import JsonSchemaException as _JsonSchemaException
from fastjsonschema import compile as _compile

# internal
from vcorelib.io import ARBITER as _ARBITER
from vcorelib.io.types import JsonObject as _JsonObject
from vcorelib.paths import find_file
from vcorelib.schemas.base import Schema as _Schema
from vcorelib.schemas.base import SchemaMap as _SchemaMap
from vcorelib.schemas.base import SchemaValidationError

LOG = _getLogger(__name__)


def package_handler(uri: str) -> _JsonObject:
    """Load data from a package."""

    parsed = _urlparse(uri)
    assert parsed.scheme == "package"

    path = find_file(parsed.path[1:], package=parsed.hostname, logger=LOG)
    assert path is not None, path
    return _ARBITER.decode(path, require_success=True).data


class JsonSchema(_Schema):
    """
    An object wrapper for: https://horejsek.github.io/python-fastjsonschema/.

    See also: https://json-schema.org/.
    """

    def __init__(self, data: _JsonObject, **_) -> None:
        """Initialize this schema."""
        super().__init__(data)
        self.validator = _compile(data, handlers={"package": package_handler})

    def __call__(self, data: _Any) -> _Any:
        """Validate input data and return the result."""
        try:
            return self.validator(data)
        except _JsonSchemaException as exc:
            raise SchemaValidationError(exc) from exc


class JsonSchemaMap(_SchemaMap):
    """A class for managing multiple schema objects."""

    @classmethod
    def kind(cls) -> _Type[JsonSchema]:
        """Implement this to determine the concrete schema type."""
        return JsonSchema



"""
A module implementing a task that can be used to sleep.
"""

# built-in
import asyncio

# internal
from vcorelib.task import Inbox as _Inbox
from vcorelib.task import Outbox, Task


class SleepTask(Task):
    """A task for sleeping for some duration."""

    async def run(
        self, inbox: _Inbox, outbox: Outbox, *args, **kwargs
    ) -> bool:
        """Sleep for the amount of time specified."""

        # Allow duration to be specified by keyword argument or positional.
        keyword = {**kwargs}
        duration = keyword.get("duration", args[0])
        await asyncio.sleep(duration)
        return True



"""
A module for implementing namespaces.
"""

# built-in
from contextlib import contextmanager as _contextmanager
from re import compile as _compile
from typing import Iterator as _Iterator
from typing import List as _List
from typing import Set as _Set

DEFAULT_DELIM = "."


class Namespace:
    """A class for implementing a basic namespace interface."""

    def __init__(self, *names: str, delim: str = DEFAULT_DELIM) -> None:
        """Initialize this namespace."""

        self.stack: _List[str] = [*names]
        self.names: _Set[str] = set()
        self.delim = delim

        # Use this attribute from preventing this namespace from colliding with
        # its parent namespace.
        self.root_size = len(self.stack)

    def child(self, *names: str) -> "Namespace":
        """Create a child namespace from this one."""
        return Namespace(*self.stack, *names, delim=self.delim)

    def push(self, name: str) -> None:
        """Push a name onto the stack."""
        self.stack.append(name)

    def pop(self, name: str = None) -> str:
        """Pop the latest name off the stack."""

        # Ensure that the namespace can't be traversed beyond its
        # initialization path.
        curr_size = len(self.stack)
        if curr_size <= self.root_size:
            raise IndexError(
                (
                    f"Root namespace has {self.root_size} and this "
                    f"namespace currently has {curr_size}!"
                )
            )

        val = self.stack.pop()

        assert (
            val == name if name is not None else True
        ), f"'{val}' != '{name}'!"
        return val

    @_contextmanager
    def pushed(self, *names: str) -> _Iterator[None]:
        """
        Provide this namespace with some names pushed onto the stack as a
        context.
        """

        to_push = [*names]
        for name in to_push:
            self.push(name)

        yield

        for name in reversed(to_push):
            self.pop(name=name)

    def namespace(
        self, name: str = None, delim: str = None, track: bool = True
    ) -> str:
        """
        Get the current namespace string with or without an additional name
        applied.
        """
        if delim is None:
            delim = self.delim

        # Skip empty names.
        result = delim.join(x for x in self.stack if x)

        if name is not None:
            if result:
                result += delim
            result += name

        # Keep track of all names added to this namespace.
        if track:
            self.names.add(result)
        return result

    def search(self, *names: str, pattern: str = ".*") -> _Iterator[str]:
        """
        Iterate over names in this namespace that match a given pattern.
        """

        # Push provided names onto the stack while yielding matches.
        with self.pushed(*names):
            start = self.namespace(track=False)

            # Add a trailing delimeter if we land on a non-empty name. Also
            # enforce that the namespaced portion is at the beginning.
            if start:
                start = "^" + start + self.delim

            # Ensure that dots are escaped to match literally.
            if start and self.delim == ".":
                start = start.replace(".", "\\.")

            # Allow the provided search string to appear anywhere in the name.
            if start:
                start += ".*"

            compiled = _compile(start + pattern)
            for name in self.names:
                if compiled.search(name) is not None:
                    yield name


class NamespaceMixin:
    """A class for giving arbitrary objects namespace capabilities."""

    def __init__(
        self, namespace: Namespace = None, namespace_delim: str = DEFAULT_DELIM
    ) -> None:
        """Initialize a namespace for this object."""

        if not hasattr(self, "_namespace"):
            if namespace is None:
                namespace = Namespace(delim=namespace_delim)
            self._namespace = namespace

    def _normalize_namespace(self, namespace: Namespace = None) -> Namespace:
        """Return this instance's namespace if one isn't provided."""

        if namespace is None:
            namespace = self._namespace
        return namespace

    def namespace(
        self, name: str = None, delim: str = None, namespace: Namespace = None
    ) -> str:
        """Get a namespace string for this object."""

        return self._normalize_namespace(namespace=namespace).namespace(
            name=name, delim=delim
        )

    def child_namespace(
        self, *names: str, namespace: Namespace = None
    ) -> Namespace:
        """Obtain a child namespace."""

        return self._normalize_namespace(namespace=namespace).child(*names)

    @_contextmanager
    def names_pushed(
        self, *names: str, namespace: Namespace = None
    ) -> _Iterator[None]:
        """Apply some names to this object's namespace as a managed context."""

        with self._normalize_namespace(namespace=namespace).pushed(*names):
            yield

    def push_name(self, name: str, namespace: Namespace = None) -> None:
        """Push a name onto the stack."""

        self._normalize_namespace(namespace=namespace).push(name)

    def pop_name(self, name: str = None, namespace: Namespace = None) -> str:
        """Pop the latest name off the stack."""

        return self._normalize_namespace(namespace=namespace).pop(name=name)

    def namespace_search(
        self, *names: str, pattern: str = ".*", namespace: Namespace = None
    ) -> _Iterator[str]:
        """Perform a search on the namespace."""

        yield from self._normalize_namespace(namespace=namespace).search(
            *names, pattern=pattern
        )



"""
A module exposing data-file encoders and decoders.
"""

# built-in
from contextlib import contextmanager as _contextmanager
from enum import Enum
import logging
from pathlib import Path
from typing import Any as _Any
from typing import Callable as _Callable
from typing import Iterator as _Iterator
from typing import NamedTuple
from typing import Optional as _Optional
from typing import cast as _cast

# internal
from vcorelib import DEFAULT_ENCODING as _DEFAULT_ENCODING
from vcorelib.dict import GenericStrDict as _GenericStrDict
from vcorelib.dict import consume, merge
from vcorelib.io.decode import (
    decode_ini,
    decode_json,
    decode_toml,
    decode_yaml,
)
from vcorelib.io.encode import (
    encode_ini,
    encode_json,
    encode_toml,
    encode_yaml,
)
from vcorelib.io.types import DEFAULT_DATA_EXT as _DEFAULT_DATA_EXT
from vcorelib.io.types import DataDecoder as _DataDecoder
from vcorelib.io.types import DataEncoder as _DataEncoder
from vcorelib.io.types import DataStream as _DataStream
from vcorelib.io.types import EncodeResult as _EncodeResult
from vcorelib.io.types import FileExtension
from vcorelib.io.types import JsonObject as _JsonObject
from vcorelib.io.types import JsonValue as _JsonValue
from vcorelib.io.types import LoadResult
from vcorelib.io.types import StreamProcessor as _StreamProcessor
from vcorelib.paths import Pathlike as _Pathlike
from vcorelib.paths import find_file, get_file_ext, get_file_name, normalize


class _DataHandle(NamedTuple):
    """A description of a data type."""

    decoder: _DataDecoder
    encoder: _DataEncoder


class DataMapping:
    """
    Map interfaces that read and write data formats to the file extensions
    that most-likely indicate the desire for that kind of format.
    """

    class DataType(Enum):
        """An aggregation of all known data types."""

        INI = _DataHandle(decode_ini, encode_ini)
        JSON = _DataHandle(decode_json, encode_json)
        YAML = _DataHandle(decode_yaml, encode_yaml)
        TOML = _DataHandle(decode_toml, encode_toml)

    mapping = {
        FileExtension.INI: DataType.INI,
        FileExtension.JSON: DataType.JSON,
        FileExtension.YAML: DataType.YAML,
        FileExtension.TOML: DataType.TOML,
    }

    @staticmethod
    def from_ext(ext: FileExtension = None) -> _Optional["DataType"]:
        """Map a file extension to a data type."""

        if ext is None:
            ext = FileExtension.UNKNOWN
        return DataMapping.mapping.get(ext)

    @staticmethod
    def from_ext_str(ext: str) -> _Optional["DataType"]:
        """Get a data type from a file-extension string."""
        return DataMapping.from_ext(FileExtension.from_ext(ext))


class DataArbiter:
    """
    A class for looking up encode and decode functions for given data types.
    """

    def __init__(
        self,
        logger: logging.Logger = logging.getLogger(__name__),
        encoding: str = _DEFAULT_ENCODING,
    ) -> None:
        """Initialize a new data arbiter."""
        self.logger = logger
        self.encoding = encoding

    def _decoder(self, ext: str) -> _Optional[_DataDecoder]:
        """Look up a decoding routine from a file extension."""

        result = None
        dtype = DataMapping.from_ext_str(ext)
        if dtype is not None:
            result = dtype.value.decoder
        else:
            self.logger.warning("No decoder for '%s'.", ext)
        return result

    def decode_stream(
        self,
        ext: str,
        stream: _DataStream,
        logger: logging.Logger = None,
        **kwargs,
    ) -> LoadResult:
        """Attempt to load data from a text stream."""

        if logger is None:
            logger = self.logger

        result = LoadResult({}, False)
        decoder = self._decoder(ext)
        if decoder is not None:
            result = decoder(stream, logger, **kwargs)
        return result

    def decode(
        self,
        pathlike: _Pathlike,
        logger: logging.Logger = None,
        require_success: bool = False,
        includes_key: _Any = None,
        preprocessor: _StreamProcessor = None,
        maxsplit: int = 1,
        **kwargs,
    ) -> LoadResult:
        """Attempt to load data from a file."""

        result = LoadResult({}, False)
        if logger is None:
            logger = self.logger

        path = normalize(pathlike)
        if path.is_file():
            with path.open(encoding=self.encoding) as path_fd:
                # Preprocess the stream if specified. This can be useful for
                # external code to treat data files as templates.
                result = self.decode_stream(
                    get_file_ext(path, maxsplit=maxsplit),
                    preprocessor(path_fd)
                    if preprocessor is not None
                    else path_fd,
                    logger,
                    **kwargs,
                )

                # Resolve includes if necessary.
                if includes_key is not None:
                    for include in consume(result.data, includes_key, []):
                        # Load the included file.
                        result = result.merge(
                            self.decode(
                                find_file(include, relative_to=path),
                                logger,
                                require_success,
                                includes_key,
                                **kwargs,
                            )
                        )

        if not result.success:
            logger.error("Failed to decode '%s'.", path)

        # Raise an exception if we expected decoding to succeed.
        if require_success:
            result.require_success(path)

        return result

    def decode_directory(
        self,
        pathlike: _Pathlike,
        logger: logging.Logger = None,
        require_success: bool = False,
        path_filter: _Callable[[Path], bool] = None,
        recurse: bool = False,
        **kwargs,
    ) -> LoadResult:
        """
        Attempt to decode data files in a directory. Assigns data loaded from
        each file to a key, returns whether or not any files failed to load
        and the cumulative time that each file-load took.
        """

        data: _JsonObject = {}
        path = normalize(pathlike)
        errors = 0
        load_time = 0

        for child in filter(
            path_filter if path_filter is not None else lambda _: True,
            path.iterdir(),
        ):
            load = None
            if child.is_file():
                load = self.decode(child, logger, require_success, **kwargs)
            elif recurse and child.is_dir():
                load = self.decode_directory(
                    child,
                    logger,
                    require_success,
                    path_filter,
                    recurse,
                    **kwargs,
                )

            if load is not None:
                if load.success:
                    key = get_file_name(child)
                    if key:
                        data[key] = _cast(_JsonValue, load.data)
                    else:
                        data = merge(
                            data,
                            load.data,
                            expect_overwrite=kwargs.get(
                                "expect_overwrite", False
                            ),
                            logger=logger,
                        )
                errors += int(not load.success)
                if load.time_ns:
                    load_time += load.time_ns

        return LoadResult(data, errors == 0, load_time)

    def encode_stream(
        self,
        ext: str,
        stream: _DataStream,
        data: _JsonObject,
        logger: logging.Logger = None,
        **kwargs,
    ) -> _EncodeResult:
        """Serialize data to an output stream."""

        if logger is None:
            logger = self.logger

        result = False
        encoder = self._encoder(ext)
        time_ns = -1
        if encoder is not None:
            time_ns = encoder(data, stream, logger, **kwargs)
            result = True
        return result, time_ns

    def encode(
        self,
        pathlike: _Pathlike,
        data: _JsonObject,
        logger: logging.Logger = None,
        maxsplit: int = 1,
        **kwargs,
    ) -> _EncodeResult:
        """Encode data to a file on disk."""

        path = normalize(pathlike)
        with path.open("w", encoding=self.encoding) as path_fd:
            return self.encode_stream(
                get_file_ext(path, maxsplit=maxsplit),
                path_fd,
                data,
                logger,
                **kwargs,
            )

    def encode_directory(
        self,
        pathlike: _Pathlike,
        data: _JsonObject,
        ext: str = _DEFAULT_DATA_EXT,
        logger: logging.Logger = None,
        **kwargs,
    ) -> _EncodeResult:
        """
        Encode data to a directory where every key becomes a file with the
        provided extension. The encoding scheme is implied by the extension.
        """

        success = True
        time_ns = 0
        root = normalize(pathlike)

        for key, item in data.items():
            result = self.encode(
                root.joinpath(f"{key}.{ext}"),
                _cast(_JsonObject, item),
                logger,
                **kwargs,
            )
            success &= result[0]
            if result[1]:
                time_ns += result[1]

        return success, time_ns

    def _encoder(self, ext: str) -> _Optional[_DataEncoder]:
        """Look up an encoding routine from a file extension."""

        result = None
        dtype = DataMapping.from_ext_str(ext)
        if dtype is not None:
            result = dtype.value.encoder
        else:
            self.logger.warning("No encoder for '%s'.", ext)
        return result

    @_contextmanager
    def object_file_context(
        self,
        pathlike: _Pathlike,
        decode_kwargs: _GenericStrDict = None,
        encode_kwargs: _GenericStrDict = None,
    ) -> _Iterator[_JsonObject]:
        """
        Provide data loaded from a file as a context so that it's written back
        when the context ends.
        """
        if decode_kwargs is None:
            decode_kwargs = {}
        if encode_kwargs is None:
            encode_kwargs = {}
        data = self.decode(pathlike, **decode_kwargs).data
        yield data
        self.encode(pathlike, data, **encode_kwargs)

    @_contextmanager
    def object_directory_context(
        self,
        pathlike: _Pathlike,
        decode_kwargs: _GenericStrDict = None,
        encode_kwargs: _GenericStrDict = None,
    ) -> _Iterator[_JsonObject]:
        """Provide a loaded directory as a context."""

        if decode_kwargs is None:
            decode_kwargs = {}
        if encode_kwargs is None:
            encode_kwargs = {}

        # Make sure the directory exists, since we would create it anyway.
        path = normalize(pathlike)
        path.mkdir(parents=True, exist_ok=True)

        data = self.decode_directory(path, **decode_kwargs).data
        yield data
        self.encode_directory(
            pathlike,
            data,
            **encode_kwargs,
        )


ARBITER = DataArbiter()



"""
A module for implementing classes that can be loaded from dictionaries and
exported as dictionaries.
"""

# built-in
import abc as _abc
from typing import Optional as _Optional
from typing import Type as _Type
from typing import TypeVar as _TypeVar

# internal
from vcorelib.io import ARBITER as _ARBITER
from vcorelib.io import DataArbiter as _DataArbiter
from vcorelib.io.types import EncodeResult as _EncodeResult
from vcorelib.io.types import JsonObject as _JsonObject
from vcorelib.paths import Pathlike as _Pathlike
from vcorelib.schemas.base import SchemaMap as _SchemaMap
from vcorelib.schemas.mixins import SchemaMixin

T = _TypeVar("T", bound="JsonCodec")


class JsonCodec(_abc.ABC, SchemaMixin):
    """A simple JSON codec interface."""

    @classmethod
    @_abc.abstractmethod
    def create(
        cls: _Type[T],
        data: _JsonObject = None,
        schemas: _SchemaMap = None,
        dest_attr: str = "data",
        verify: bool = True,
    ) -> T:
        """A factory method to create an instance."""

    def __copy__(self: T) -> T:
        """Create another instance based on this instance's data."""
        return self.__class__.create(self.asdict())

    def __str__(self) -> str:
        """
        Use the dictionary representation of this instance for string
        representation.
        """
        return str(self.asdict())

    def __eq__(self, other) -> bool:
        """Determine if this instance is equal to another."""

        # Allow direct comparison with another dictionary, but not an arbitrary
        # mapping.
        if isinstance(other, dict):
            to_cmp = other
        else:
            to_cmp = other.asdict()

        return self.asdict() == to_cmp

    @_abc.abstractmethod
    def asdict(self) -> _JsonObject:
        """Obtain a dictionary representing this instance."""

    def encode(
        self, pathlike: _Pathlike, arbiter: _DataArbiter = _ARBITER, **kwargs
    ) -> _EncodeResult:
        """Encode this object instance to a file."""
        return arbiter.encode(pathlike, self.asdict(), **kwargs)

    @classmethod
    def decode(
        cls: _Type[T],
        pathlike: _Pathlike,
        arbiter: _DataArbiter = _ARBITER,
        schemas: _SchemaMap = None,
        dest_attr: str = "data",
        verify: bool = True,
        **kwargs,
    ) -> T:
        """Decode an object instance from data loaded from a file."""

        return cls.create(
            arbiter.decode(pathlike, require_success=True, **kwargs).data,
            schemas=schemas,
            dest_attr=dest_attr,
            verify=verify,
        )


V = _TypeVar("V", bound="DictCodec")


class DictCodec(JsonCodec):
    """
    A class implementing an interface for objects that can be decoded from
    disk and encoded back to disk.
    """

    default_schemas: _Optional[_SchemaMap] = None

    def __init__(
        self,
        data: _JsonObject = None,
        schemas: _SchemaMap = None,
        dest_attr: str = "data",
        verify: bool = True,
    ) -> None:
        """Initialize this instance."""

        if data is None:
            data = {}
        setattr(self, dest_attr, data)

        # Validate the instance's dictionary data based on a schema.
        if schemas is None:
            schemas = self.default_schemas
        if schemas is not None:
            super().__init__(schemas, valid_attr=dest_attr)

        self.init(getattr(self, dest_attr))

        # After initialization, creating a new dictionary from this object
        # should be equivalent to the one provided to the constructor. This
        # can be disabled for special cases.
        if verify:
            assert self == data, f"'{self}' != '{data}' after initialization!"

    @_abc.abstractmethod
    def init(self, data: _JsonObject) -> None:
        """Perform implementation-specific initialization."""

    @classmethod
    def create(
        cls: _Type[V],
        data: _JsonObject = None,
        schemas: _SchemaMap = None,
        dest_attr: str = "data",
        verify: bool = True,
    ) -> V:
        """A factory method to create an instance."""
        return cls(data, schemas=schemas, dest_attr=dest_attr, verify=verify)


class BasicDictCodec(DictCodec):
    """The simplest possible dictionary codec implementation."""

    def init(self, data: _JsonObject) -> None:
        """Initialize this instance."""
        self.data = data

    def asdict(self) -> _JsonObject:
        """Obtain a dictionary representing this instance."""
        return self.data.copy()



"""
A module for abstract base-classes relevant to IO.
"""

# built-in
import abc as _abc
from typing import TextIO as _TextIO

# internal
from vcorelib import DEFAULT_ENCODING as _DEFAULT_ENCODING
from vcorelib.paths import Pathlike as _Pathlike
from vcorelib.paths import normalize as _normalize


class Serializable(_abc.ABC):
    """An interface definition for serializable classes."""

    def default_location(self) -> _Pathlike:
        """Get a default serialization destination for this instance."""
        return f"{self.__class__.__name__}.txt"

    @_abc.abstractmethod
    def to_stream(self, stream: _TextIO, **kwargs) -> None:
        """Write this object to a text stream."""

    def to_file(
        self,
        path: _Pathlike = None,
        encoding: str = _DEFAULT_ENCODING,
        **kwargs,
    ) -> None:
        """Write this object to a file."""

        with _normalize(
            path if path is not None else self.default_location()
        ).open("w", encoding=encoding) as stream:
            self.to_stream(stream, **kwargs)
            stream.flush()


class FileEntity(Serializable):
    """A class for working with objects that are backed by a disk location."""

    def __hash__(self) -> int:
        """Use this instance's disk location for hashing."""
        return hash(self.location)

    def default_location(self) -> _Pathlike:
        """Get a default serialization destination for this instance."""
        return self.location

    def __init__(self, location: _Pathlike) -> None:
        """Initialize this file-backed entity."""
        assert not hasattr(self, "location")
        self.location = _normalize(location)



"""
Utilities for working with environment-variable data contained inside various
data structures.
"""

# built-in
from os import environ as _environ
from typing import Any as _Any
from typing import List as _List

# internal
from vcorelib.dict import GenericDict as _GenericDict

GenericList = _List[_Any]


def str_resolve_env_var(data: str) -> str:
    """
    Convert string data to a resolved environment variable if the string begins
    with '$' and is a key in the environment with a non-empty value.
    """

    temp = data.strip()
    if temp.startswith("$"):
        key = temp[1:]
        if key in _environ and _environ[key]:
            data = _environ[key]

    return data


def list_resolve_env_vars(
    data: GenericList,
    keys: bool = True,
    values: bool = True,
    lists: bool = True,
) -> GenericList:
    """
    Recursively resolve list data that may contain strings that should be
    treated as environment-variable substitutions. The data is updated
    in-place.
    """

    for idx, item in enumerate(data):
        if isinstance(item, dict) and (keys or values):
            data[idx] = dict_resolve_env_vars(item, keys, values, lists)
        if isinstance(item, list) and lists:
            data[idx] = list_resolve_env_vars(item, keys, values, lists)
        if isinstance(item, str):
            data[idx] = str_resolve_env_var(item)

    return data


def dict_resolve_env_vars(
    data: _GenericDict,
    keys: bool = True,
    values: bool = True,
    lists: bool = True,
) -> _GenericDict:
    """
    Recursively resolve dictionary data that may contain strings that should be
    treated as environment-variable substitutions. The data is updated
    in-place.
    """

    keys_to_remove: _List[str] = []
    to_update: _GenericDict = {}
    for key, value in data.items():
        if isinstance(value, str) and values:
            value = str_resolve_env_var(value)

        if isinstance(key, str) and keys:
            to_update[str_resolve_env_var(key)] = value
            keys_to_remove.append(key)

        data[key] = value

        if isinstance(value, dict):
            data[key] = dict_resolve_env_vars(value, keys, values, lists)
        if isinstance(value, list) and lists:
            data[key] = list_resolve_env_vars(value, keys, values, lists)

    # Don't change the size of data while iterating.
    for key in keys_to_remove:
        del data[key]
    data.update(to_update)

    return data



# =====================================
# generator=datazen
# version=3.1.0
# hash=667a45b6e3039f998be6e2945ea7cc28
# =====================================

"""
Useful defaults and other package metadata.
"""

DESCRIPTION = "A collection of core Python utilities."
PKG_NAME = "vcorelib"
VERSION = "1.2.5"

# vcorelib-specific content.
DEFAULT_INDENT = 2
DEFAULT_ENCODING = "utf-8"



"""
A module for implementing graph-like data structures.

See Graphviz's 'DOT Language' here: https://graphviz.org/doc/info/lang.html.
"""

# built-in
from os import linesep as _linesep
from typing import TextIO as _TextIO

# internal
from vcorelib import DEFAULT_INDENT as _DEFAULT_INDENT
from vcorelib.graph.abc import AbstractDiGraph as _AbstractDiGraph
from vcorelib.graph.edge import write_attributes as _write_attributes


def write_indent(
    indent: int, stream: _TextIO, linesep: str = _linesep
) -> None:
    """Handle indented prefixes for new lines in a text stream."""

    if indent > 0:
        if linesep:
            stream.write(linesep)
        stream.write(" " * indent)


class DiGraph(_AbstractDiGraph):
    """A simple, directed-graph implementation."""

    def to_stream(self, stream: _TextIO, **kwargs) -> None:
        """Write this object to a text stream."""

        indent: int = kwargs.pop("indent", _DEFAULT_INDENT)
        linesep: str = kwargs.pop("linesep", _linesep)

        stream.write(f"strict digraph {self.name} {{")

        # Add top-level graph attributes.
        if self.graph_attrs:
            write_indent(indent, stream, linesep=linesep)
            stream.write("graph")
            _write_attributes(stream, data=self.graph_attrs)
            stream.write(";")

        # Add top-level node attributes.
        if self.node_attrs:
            write_indent(indent, stream, linesep=linesep)
            stream.write("node")
            _write_attributes(stream, data=self.node_attrs)
            stream.write(";")

        # Add top-level edge attributes.
        if self.edge_attrs:
            write_indent(indent, stream, linesep=linesep)
            stream.write("edge")
            _write_attributes(stream, data=self.edge_attrs)
            stream.write(";")

        # Write nodes.
        for node in self.data.values():
            write_indent(indent, stream, linesep=linesep)
            node.to_stream(stream)

        # Write edges.
        for edge_set in self.edges.values():
            for edge in edge_set:
                write_indent(indent, stream, linesep=linesep)
                edge.to_stream(stream)

        stream.write(linesep if indent > 0 else " ")
        stream.write("}")



"""
A module defining interfaces for schema enforcement.
"""

# built-in
import abc as _abc
from collections import UserDict
from typing import Any as _Any
from typing import Iterator as _Iterator
from typing import MutableMapping as _MutableMapping
from typing import Tuple as _Tuple
from typing import Type as _Type
from typing import TypeVar as _TypeVar

# internal
from vcorelib.io import ARBITER as _ARBITER
from vcorelib.io.types import JsonObject as _JsonObject
from vcorelib.paths import Pathlike as _Pathlike
from vcorelib.paths import get_file_name as _get_file_name
from vcorelib.paths import normalize as _normalize
from vcorelib.paths import resource as _resource

T = _TypeVar("T", bound="Schema")


class SchemaValidationError(Exception):
    """An exception type for schema errors."""


class Schema(_abc.ABC):
    """A base class for schema enforcement."""

    def __init__(self, data: _JsonObject, **_) -> None:
        """Initialize this schema."""
        self.data = data

    @_abc.abstractmethod
    def __call__(self, data: _Any) -> _Any:
        """Validate input data and return the result."""

    @classmethod
    def from_path(cls: _Type[T], path: _Pathlike, **kwargs) -> T:
        """Load a schema from a data file on disk."""
        return cls(_ARBITER.decode(path, require_success=True).data, **kwargs)


V = _TypeVar("V", bound="SchemaMap")


class SchemaMap(
    UserDict,  # type: ignore
    _MutableMapping[str, Schema],
):
    """A class for managing multiple schema objects."""

    @classmethod
    @_abc.abstractmethod
    def kind(cls: _Type[V]) -> _Type[Schema]:
        """Implement this to determine the concrete schema type."""

    def __init__(self) -> None:
        """Initialize this schema map."""
        super().__init__(self)

    def load_file(self, path: _Pathlike, **kwargs) -> _Tuple[str, T]:
        """Load a schema file into the map."""

        path = _normalize(path)
        name = _get_file_name(path)
        assert name not in self, f"Duplicate schema '{name}'!"
        self[name] = self.kind().from_path(path, **kwargs)
        return name, self[name]

    def load_directory(
        self, path: _Pathlike, **kwargs
    ) -> _Iterator[_Tuple[str, Schema]]:
        """Load a directory of schema files into the map."""

        path = _normalize(path)
        assert path.is_dir(), f"'{path}' isn't a directory!"
        for item in path.iterdir():
            yield self.load_file(item, **kwargs)

    def load_package(
        self,
        package: str,
        path: _Pathlike = "schemas",
        package_subdir: str = "data",
        **kwargs,
    ) -> _Iterator[_Tuple[str, Schema]]:
        """Load schemas from package data."""

        path = _resource(path, package=package, package_subdir=package_subdir)
        assert (
            path is not None and path.is_dir()
        ), f"Can't find schema directory for package '{package}'!"

        yield from self.load_directory(path, **kwargs)

    @classmethod
    def from_package(
        cls: _Type[V],
        package: str,
        **kwargs,
    ) -> V:
        """Create a new JSON-schema map from package data."""

        result = cls()
        list(result.load_package(package, **kwargs))
        return result



"""
A module implementing a JSON-file database.
"""

# built-in
from abc import ABC, abstractmethod
from contextlib import ExitStack as _ExitStack
from contextlib import contextmanager as _contextmanager
from typing import Iterator as _Iterator

# internal
from vcorelib.io import ARBITER as _ARBITER
from vcorelib.io import DataArbiter as _DataArbiter
from vcorelib.io.types import JsonObject as _JsonObject
from vcorelib.paths import Pathlike as _Pathlike
from vcorelib.paths import normalize as _normalize


class JsonCache(ABC):
    """A JSON cach interface."""

    def __init__(
        self, pathlike: _Pathlike, arbiter: _DataArbiter = _ARBITER
    ) -> None:
        """Initialize this cache."""
        self.path = _normalize(pathlike)
        self.arbiter = arbiter

    @abstractmethod
    def context_load(self, stack: _ExitStack, **kwargs) -> _JsonObject:
        """Load the data."""

    @_contextmanager
    def loaded(self, **kwargs) -> _Iterator[_JsonObject]:
        """
        Provide loaded data so that the data is written back to disk on
        completion.
        """
        with _ExitStack() as stack:
            yield self.context_load(stack, **kwargs)


class FileCache(JsonCache):
    """A class implementing a JSON cache based on a file."""

    def context_load(self, stack: _ExitStack, **kwargs) -> _JsonObject:
        """Load the data."""
        return stack.enter_context(
            self.arbiter.object_file_context(self.path, **kwargs)
        )


class DirectoryCache(JsonCache):
    """A class implementing a JSON cache based on a directory."""

    def context_load(self, stack: _ExitStack, **kwargs) -> _JsonObject:
        """Load the data."""
        return stack.enter_context(
            self.arbiter.object_directory_context(self.path, **kwargs)
        )



"""
Dictionary manipulation utilities.
"""

from contextlib import contextmanager as _contextmanager

# built-in
from enum import Enum as _Enum
from enum import auto as _auto
from logging import Logger, getLogger
from typing import Any as _Any
from typing import Dict as _Dict
from typing import Iterator as _Iterator
from typing import List as _List

_LOG = getLogger(__name__)
GenericDict = _Dict[_Any, _Any]
GenericStrDict = _Dict[str, _Any]


def consume(data: GenericDict, key: _Any, default: _Any = None) -> _Any:
    """
    Attempt to obtain dictionary data via key, removing the data if it was
    present.
    """
    return data.pop(key, default)


def set_if_not(data: GenericDict, key: _Any, value: _Any = None) -> _Any:
    """
    Set a value in a dictionary if one wasn't already set and return the value
    that ends up at that key.
    """

    if key not in data:
        data[key] = value
    return data[key]


@_contextmanager
def limited(
    data: GenericDict, key: _Any, value: _Any = None
) -> _Iterator[None]:
    """Ensure that dictionary data is only temporarily added."""

    had_key = False
    orig_value = None

    # Provide the new value.
    if value is not None:
        had_key = key in data
        if had_key:
            orig_value = data[key]
        data[key] = value

    # If no value is provided, ensure that this key isn't already present
    # to prevent ambiguity.
    else:
        assert key not in data

    yield

    # Restore the dictionary to its initial state.
    if value is not None:
        del data[key]
        if had_key:
            data[key] = orig_value


class MergeStrategy(_Enum):
    """
    An enumeration describing strategies for combining various data structures.
    """

    RECURSIVE = _auto()
    UPDATE = _auto()


def merge_recursive(
    dict_a: GenericDict,
    dict_b: GenericDict,
    path: _List[str] = None,
    expect_overwrite: bool = False,
    logger: Logger = None,
) -> GenericDict:
    """
    Combine two dictionaries recursively, prefers dict_a in a conflict. For
    values of the same key that are lists, the lists are combined. Otherwise
    the resulting dictionary is cleanly merged.
    """

    if path is None:
        path = []
    if logger is None:
        logger = _LOG

    for key, right_val in dict_b.items():
        if key not in dict_a:
            dict_a[key] = right_val
            continue

        # first try to coerce b's type into a's
        if not isinstance(right_val, type(dict_a[key])):
            try:
                right_val = type(dict_a[key])(right_val)
            except ValueError:
                pass

        # same leaf value
        if dict_a[key] == right_val:
            pass
        elif isinstance(dict_a[key], dict) and isinstance(right_val, dict):
            merge(
                dict_a[key],
                right_val,
                path + [str(key)],
                expect_overwrite,
                logger,
            )
        elif isinstance(dict_a[key], list) and isinstance(right_val, list):
            dict_a[key].extend(right_val)
        elif not isinstance(right_val, type(dict_a[key])):
            logger.error("Type mismatch at '%s'", ".".join(path + [str(key)]))
            logger.error("left:  %s (%s)", type(dict_a[key]), dict_a[key])
            logger.error("right: %s (%s)", type(right_val), right_val)
        elif not expect_overwrite:
            logger.error("Conflict at '%s'", ".".join(path + [str(key)]))
            logger.error("left:  %s", dict_a[key])
            logger.error("right: %s", right_val)
        else:
            dict_a[key] = right_val

    return dict_a


def merge(
    dict_a: GenericDict,
    dict_b: GenericDict,
    path: _List[str] = None,
    expect_overwrite: bool = False,
    logger: Logger = None,
    strategy: MergeStrategy = MergeStrategy.RECURSIVE,
) -> GenericDict:
    """Combine two dictionaries based on a provided merge strategy."""

    if strategy is MergeStrategy.UPDATE:
        dict_a.update(dict_b)
        return dict_a

    return merge_recursive(
        dict_a,
        dict_b,
        path=path,
        expect_overwrite=expect_overwrite,
        logger=logger,
    )


def merge_dicts(
    dicts: _List[GenericDict],
    expect_overwrite: bool = False,
    logger: Logger = None,
    strategy: MergeStrategy = MergeStrategy.RECURSIVE,
) -> GenericDict:
    """
    Merge a list of dictionary data into a single set (mutates the first
    element).
    """

    result = dicts[0]
    for right_dict in dicts[1:]:
        result = merge(
            result,
            right_dict,
            expect_overwrite=expect_overwrite,
            logger=logger,
            strategy=strategy,
        )
    return result



"""
Test the 'namespace' module.
"""

# third-party
from pytest import raises

# module under test
from vcorelib.namespace import Namespace, NamespaceMixin


def test_namespace_basic():
    """Test basic functionality of namespaces."""

    class NamespaceEntity(NamespaceMixin):
        """A sample class."""

    inst = NamespaceEntity()

    child = inst.child_namespace("test")
    with raises(IndexError):
        child.pop()

    with inst.names_pushed("a", "b", "c"):
        assert inst.namespace() == "a.b.c"
        assert inst.namespace("d") == "a.b.c.d"
    assert inst.namespace("d") == "d"
    inst.push_name("a")
    assert inst.pop_name("a") == "a"

    assert list(inst.namespace_search("a", "b", "c", pattern="d")) == [
        "a.b.c.d"
    ]


def test_namespace_search():
    """Test that we can search namespaces for names."""

    names = Namespace()

    sample = [f"name_{x}" for x in "abc"]
    for item in sample:
        names.namespace(name=item)

    assert sorted(list(names.search())) == sample
    assert sorted(list(names.search(pattern="name"))) == sample
    assert list(names.search(pattern="_a")) == ["name_a"]

    # Add names to a sepecific namespace.
    with names.pushed("test"):
        for item in sample:
            names.namespace(name=item)

    assert sorted(list(names.search(pattern="test"))) == [
        "test.name_a",
        "test.name_b",
        "test.name_c",
    ]

    assert sorted(list(names.search("test", pattern="name"))) == [
        "test.name_a",
        "test.name_b",
        "test.name_c",
    ]

    assert list(names.search("test", pattern="b")) == ["test.name_b"]



"""
Math utilities.
"""

# built-in
import typing


class UnitSystem(typing.NamedTuple):
    """
    A pairing of prefixes defining a unit, and the amount that indicates the
    multiplicative step-size between them.
    """

    prefixes: typing.Sequence[str]
    divisor: int


SI_UNITS = UnitSystem(["n", "u", "m", "", "k", "M", "G", "T"], 1000)
KIBI_UNITS = UnitSystem(
    ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi", "Yi"], 1024
)


def unit_traverse(
    val: int,
    unit: UnitSystem = SI_UNITS,
    max_prefix: int = 3,
    iteration: int = 0,
) -> typing.Tuple[int, int, str]:
    """
    Given an initial value, traverse a unit system to get the largest
    representative unit prefix. Also return a fractional component, in units
    of the next-smallest prefix.
    """

    prefixes = unit.prefixes
    divisor = unit.divisor
    decimal = val
    fractional = 0

    max_iter = min(len(prefixes) - 1, max_prefix)
    while decimal >= divisor and iteration < max_iter:
        fractional = decimal % divisor
        decimal = decimal // divisor
        iteration += 1

    return decimal, fractional, prefixes[iteration]



"""
A module for extracting and creating arbitrary archives.
"""

# built-in
from os import chdir as _chdir
from os import getcwd as _getcwd
from os import path as _path
from pathlib import Path
import shutil
import tarfile
from typing import Optional as _Optional
from typing import Tuple as _Tuple
import zipfile

# internal
from vcorelib.io.types import DEFAULT_ARCHIVE_EXT as _DEFAULT_ARCHIVE_EXT
from vcorelib.io.types import FileExtension
from vcorelib.math.time import TIMER as _TIMER
from vcorelib.paths import Pathlike as _Pathlike
from vcorelib.paths import normalize as _normalize


def is_within_directory(directory: _Pathlike, target: _Pathlike) -> bool:
    """Determine if a target path is within the provided directory."""

    directory = str(_normalize(directory).resolve())
    return (
        _path.commonprefix([directory, str(_normalize(target).resolve())])
        == directory
    )


def safe_extract(
    tar: tarfile.TarFile, path: _Pathlike = None, **kwargs
) -> None:
    """Sanity check that all members will fall within the destination path."""

    path = _normalize(path).resolve()

    for member in tar.getmembers():
        dest = path.joinpath(member.name)
        if not is_within_directory(path, dest):  # pragma: no cover
            raise OSError(
                (
                    "Attempted path traversal in tar "
                    f"file: '{dest}' not in '{path}'."
                )
            )

    # Perform the extraction.
    tar.extractall(path, **kwargs)


def extractall(
    src: _Pathlike, dst: _Pathlike = None, maxsplit: int = 1, **extract_kwargs
) -> _Tuple[bool, int]:
    """
    Attempt to extract an arbitrary archive to a destination. Return whether or
    not this succeeded and how long it took.
    """

    success = False
    time_ns = -1
    src = _normalize(src)
    ext = FileExtension.from_path(src, maxsplit=maxsplit)

    # Ensure that the source directory is an archive.
    if ext is None or not ext.is_archive() or not src.is_file():
        return success, time_ns

    dst = _normalize(dst)

    with _TIMER.measure_ns() as token:
        # Extract the tar archive.
        if ext is FileExtension.TAR:
            with tarfile.open(src) as tar:
                safe_extract(tar, dst, **extract_kwargs)
            success = True

        # Extract the ZIP archive.
        elif ext is FileExtension.ZIP:
            with zipfile.ZipFile(src) as zipf:
                zipf.extractall(dst, **extract_kwargs)
            success = True

    return success, _TIMER.result(token)


def make_archive(
    src_dir: Path,
    ext_str: str = _DEFAULT_ARCHIVE_EXT,
    dst_dir: Path = None,
    **archive_kwargs,
) -> _Tuple[_Optional[Path], int]:
    """
    Create an archive from a source directory, named after that directory,
    and optionally moved to a destination other than the parent directory
    for the source. The extension specifies the kind of archive to create.

    Return the path to the created archive (if it was created) and how long
    it took to create.
    """

    result = None
    time_ns = -1

    if not src_dir.is_dir():
        return result, time_ns

    # Map file extensions to the archiver's known formats. Some extensions
    # don't require any mapping (e.g. "tar", "zip").
    format_map = {
        "tar.gz": "gztar",
        "tar.bz2": "bztar",
        "tar.lzma": "xztar",
        "tar.xz": "xztar",
    }
    format_str = format_map.get(ext_str, ext_str)

    # Make sure that this output format is supported.
    if format_str not in [x[0] for x in shutil.get_archive_formats()]:
        return result, time_ns

    curr = _getcwd()
    try:
        _chdir(src_dir.parent)

        with _TIMER.measure_ns() as token:
            result = Path(
                shutil.make_archive(
                    src_dir.name,
                    format_str,
                    base_dir=src_dir.name,
                    **archive_kwargs,
                )
            ).resolve()
        time_ns = _TIMER.result(token)

        # Move the resulting archive, if requested.
        if dst_dir is not None:
            dst_dir.mkdir(parents=True, exist_ok=True)
            new_path = Path(dst_dir, result.name)
            assert not new_path.is_dir()
            shutil.move(str(result), new_path)
            result = new_path
    finally:
        _chdir(curr)

    return result, time_ns



"""
A task definition for melding inbox data into outbox data.
"""

# built-in
from typing import Any as _Any
from typing import Dict as _Dict
from typing import List as _List

# internal
from vcorelib.dict import merge, merge_dicts
from vcorelib.task import Inbox as _Inbox
from vcorelib.task import Outbox, Task


class DictMerger(Task):
    """A class that pipes inbox data through to outbox data."""

    async def run(
        self, inbox: _Inbox, outbox: Outbox, *args, **kwargs
    ) -> bool:
        """Override this method to implement the task."""

        # Forward the foreign argument to the ouput box (provided by the
        # caller).
        to_merge: _List[_Dict[_Any, _Any]] = [outbox]
        for arg in [*args]:
            if isinstance(arg, dict):
                to_merge.append(arg)
        to_merge.append(kwargs)

        expect_overwrite = kwargs.pop("expect_overwrite", False)
        merge_dicts(
            to_merge, expect_overwrite=expect_overwrite, logger=self.log
        )

        # Forward all inputs to the output box.
        merge(
            outbox, inbox, expect_overwrite=expect_overwrite, logger=self.log
        )
        return True



"""
vcorelib - Test the 'logging' module.
"""

# internal
from logging import getLogger

# module under test
from vcorelib.logging import LoggerMixin, log_time


def test_log_time_basic():
    """Test that log_time works in a simple scenario."""

    log = getLogger(__name__)
    with log_time(log, "Example"):
        for _ in range(100):
            pass


def test_logger_mixin_basic():
    """Test basic functionality of the logger mixin."""

    class LoggerMixinTest(
        LoggerMixin
    ):  # pylint: disable=too-few-public-methods
        """A test class."""

    inst = LoggerMixinTest()
    inst.logger.info("This is a test, %d %d %d.", 1, 2, 3)

    with inst.log_time("Hello, %s! %d", "world", 5):
        for _ in range(100):
            pass



"""
A module for implementing a configuration data manager with a dictionary.
"""

# built-in
from collections import UserDict as _UserDict
from typing import Any as _Any
from typing import MutableMapping as _MutableMapping

# internal
from vcorelib.dict import GenericDict as _GenericDict
from vcorelib.dict import consume as _consume
from vcorelib.dict import merge as _merge
from vcorelib.dict import set_if_not as _set_if_not
from vcorelib.io import ARBITER as _ARBITER
from vcorelib.io.types import FileExtension as _FileExtension
from vcorelib.paths import Pathlike as _Pathlike
from vcorelib.paths import normalize as _normalize

ConfigData = _MutableMapping[str, _Any]


class Config(
    _UserDict,  # type: ignore
    ConfigData,
):
    """A dictionary that allows access to data only once via each key."""

    def __getitem__(self, key) -> _Any:
        """Consume keys after their data is retreived."""
        assert key in self.data, f"Key '{key}' not found in configuration!"
        return _consume(self.data, key)

    def get(self, key, default=None) -> _Any:
        """Get data from the configuration but allow a default value."""
        return _consume(self.data, key, default)

    def set_if_not(self, key, value) -> _Any:
        """Set this value if a value is not already set."""
        return _set_if_not(self.data, key, value)

    def merge(self, other: _GenericDict, *args, **kwargs) -> None:
        """Merge another dictionary into this one."""
        _merge(self.data, other, *args, **kwargs)

    @staticmethod
    def from_file(pathlike: _Pathlike, *args, **kwargs) -> "Config":
        """Load a configuration from a file on disk."""
        _set_if_not(kwargs, "require_success", True)
        return Config(_ARBITER.decode(pathlike, *args, **kwargs).data)

    @staticmethod
    def from_directory(pathlike: _Pathlike, *args, **kwargs) -> "Config":
        """Load a configuration from a directory on disk."""
        _set_if_not(kwargs, "require_success", True)
        return Config(
            _ARBITER.decode_directory(pathlike, *args, **kwargs).data
        )

    @staticmethod
    def from_path(pathlike: _Pathlike, *args, **kwargs) -> "Config":
        """Load a configuration from an arbitrary path."""

        path = _normalize(pathlike)

        # Load this as a directory if it is one.
        if path.is_dir():
            return Config.from_directory(path, *args, **kwargs)

        candidates = list(
            _FileExtension.data_candidates(path, exists_only=True)
        )
        count = len(candidates)
        assert (
            count == 1
        ), f"Found {count} configurations for '{path}': {candidates}."

        # Load the data file.
        result = Config.from_file(candidates[0], *args, **kwargs)

        # Also load a directory of the stem name matches one. Load this data
        # via merging dictionaries.
        to_check = path.with_suffix("")
        if to_check.is_dir():
            result.merge(Config.from_directory(to_check).data)

        return result



"""
A module for working with graph ports.
"""

from enum import Enum as _Enum
from enum import auto as _auto

# built-in
from itertools import zip_longest as _zip_longest
from typing import Dict as _Dict
from typing import NamedTuple
from typing import Optional as _Optional
from typing import Set as _Set
from typing import TextIO as _TextIO


class PortType(_Enum):
    """An enumeration of port types."""

    INPUT = _auto()
    OUTPUT = _auto()
    INOUT = _auto()

    def __str__(self) -> str:
        """Get this port type as a string."""
        return self.name.lower()

    @property
    def is_input(self) -> bool:
        """Determine if this port type is capable of input."""
        return self is not PortType.OUTPUT

    @property
    def is_output(self) -> bool:
        """Determine if this port type is capable of output."""
        return self is not PortType.INPUT


class Port(NamedTuple):
    """A simple implementation for generic, input and output ports."""

    label: str
    kind: PortType = PortType.INOUT
    alias: _Optional[str] = None
    allocated: bool = False

    def allocate(self) -> "Port":
        """Allocate this port."""

        assert not self.allocated, f"Port '{self}' is already allocated!"
        return Port(
            self.label, kind=self.kind, alias=self.alias, allocated=True
        )

    def __str__(self) -> str:
        """Get this port as a string."""

        result = self.label
        alloc_str = "x" if self.allocated else " "
        result = f"{self.label}[{self.kind}][{alloc_str}]"
        if self.alias:
            result += f" ({self.alias})"
        return result

    @property
    def is_input(self) -> bool:
        """Determine if this is an input port."""
        return self.kind.is_input

    @property
    def is_output(self) -> bool:
        """Determine if this is an output port."""
        return self.kind.is_output


class PortManager:
    """An interface for managing input and output ports."""

    def __init__(self) -> None:
        """Initialize this port manager."""

        self.outputs: _Dict[str, Port] = {}
        self.inputs: _Dict[str, Port] = {}

    def inout_labels(self) -> _Set[str]:
        """Get inout port labels."""
        return set(self.inputs.keys()).intersection(self.outputs.keys())

    def input_labels(self, exclude_inout: bool = True) -> _Set[str]:
        """Get input port labels."""

        result = set(self.inputs.keys())
        if exclude_inout:
            result -= self.inout_labels()
        return result

    def output_labels(self, exclude_inout: bool = True) -> _Set[str]:
        """Get output port labels."""

        result = set(self.outputs.keys())
        if exclude_inout:
            result -= self.inout_labels()
        return result

    def label(self, stream: _TextIO) -> None:
        """Create a record label based on the current port configuration."""

        columns = []

        # Handle inout ports.
        inouts = "|".join(
            [f"<{x}> {self.outputs[x]}" for x in sorted(self.inout_labels())]
        )
        columns.append(inouts if inouts else "no inout ports")

        # Put inputs above outputs and arrange in a table.
        for in_label, out_label in _zip_longest(
            sorted(self.input_labels()), sorted(self.output_labels())
        ):
            in_data = (
                f"<{in_label}> {self.inputs[in_label]}" if in_label else ""
            )
            out_data = (
                f"<{out_label}> {self.outputs[out_label]}" if out_label else ""
            )
            columns.append("{" + in_data + "|" + out_data + "}")

        # Write columns to the stream.
        stream.write("|".join(columns))

    def create(self, label: str, **kwargs) -> Port:
        """Add a new port to this port manager."""

        port = Port(label, **kwargs)

        # Ensure the port's alias doesn't alias an existing input or output
        # port.
        if port.alias:
            assert (
                port.alias not in self.inputs
            ), f"Port alias '{port.alias}' is already an input label!"
            assert (
                port.alias not in self.outputs
            ), f"Port alias '{port.alias}' is already an output label!"

        if port.is_input:
            assert label not in self.inputs, f"Duplicate input port '{port}'!"
            self.inputs[label] = port

        if port.is_output:
            assert (
                label not in self.outputs
            ), f"Duplicate output port '{port}'!"
            self.outputs[label] = port

        return port

    def allocate(self, label: str) -> Port:
        """Allocate a labeled port."""

        port = None

        if label in self.inputs:
            port = self.inputs[label].allocate()
            self.inputs[label] = port

        if label in self.outputs:
            # Don't create a new port if it's an inout.
            if port is None:
                port = self.outputs[label].allocate()
            self.outputs[label] = port

        assert port is not None, f"Port '{label}' not found!"
        return port



"""
A module for working with asyncio.
"""

from __future__ import annotations

# built-in
from asyncio import AbstractEventLoop as _AbstractEventLoop
from asyncio import CancelledError as _CancelledError
from asyncio import Event as _Event
from asyncio import Task as _Task
from asyncio import all_tasks as _all_tasks
from asyncio import get_event_loop as _get_event_loop
from contextlib import suppress as _suppress
from logging import getLogger as _getLogger
from typing import Any as _Any
from typing import Awaitable as _Awaitable
from typing import Coroutine as _Coroutine
from typing import Iterable as _Iterable
from typing import List as _List
from typing import Optional as _Optional
from typing import TypeVar as _TypeVar

# internal
from vcorelib.logging import LoggerType as _LoggerType

T = _TypeVar("T")
LOG = _getLogger(__name__)


def log_task_exception(task: _Task[_Any], logger: _LoggerType = None) -> None:
    """If a task is done and raised an exception, log it."""

    if logger is None:
        logger = LOG

    if task.done():
        with _suppress(_CancelledError):
            exc = task.exception()
            if exc is not None and not isinstance(exc, _CancelledError):
                logger.exception("Task raised exception:", exc_info=exc)


def log_exceptions(
    tasks: _Iterable[_Task[T]], logger: _LoggerType = None
) -> _List[_Task[T]]:
    """Log task exception and return the list of tasks that aren't complete."""

    for task in tasks:
        log_task_exception(task, logger=logger)
    return [x for x in tasks if not x.done()]


def shutdown_loop(
    eloop: _AbstractEventLoop, logger: _LoggerType = None
) -> None:
    """Attempt to shut down an event loop."""

    eloop.run_until_complete(eloop.shutdown_asyncgens())

    tasks = log_exceptions(_all_tasks(loop=eloop), logger=logger)
    if tasks:
        # Cancel all tasks running in the event loop.
        for task in tasks:
            task.cancel()

            # Give all tasks a chance to complete.
            with _suppress(KeyboardInterrupt, _CancelledError):
                eloop.run_until_complete(task)
                log_task_exception(task, logger=logger)


def run_handle_interrupt(
    to_run: _Awaitable[_Any], eloop: _AbstractEventLoop
) -> _Optional[_Any]:
    """
    Run a task in an event loop and gracefully handle keyboard interrupts.

    Return the result of the awaitable or None if execution was interrupted.
    """

    result = None
    try:
        result = eloop.run_until_complete(to_run)
    except KeyboardInterrupt:
        shutdown_loop(eloop)

    return result


def run_handle_stop(
    stop_sig: _Event,
    task: _Coroutine[None, None, T],
    eloop: _AbstractEventLoop = None,
) -> T:
    """
    Publish the stop signal on keyboard interrupt and wait for the task to
    complete.
    """

    if eloop is None:
        eloop = _get_event_loop()
    to_run = eloop.create_task(task)

    while True:
        try:
            return eloop.run_until_complete(to_run)
        except KeyboardInterrupt:  # pragma: nocover
            print("Keyboard interrupt.")
            stop_sig.set()



"""
A module containing abstract base-classes related to graphs.
"""

# built-in
from collections import UserDict as _UserDict
from typing import Dict as _Dict
from typing import Iterator as _Iterator
from typing import Set as _Set
from typing import Type as _Type
from typing import TypeVar as _TypeVar

# internal
from vcorelib.graph.edge import AttributeMap as _AttributeMap
from vcorelib.graph.edge import GraphEdge as _GraphEdge
from vcorelib.graph.port import Port as _Port
from vcorelib.graph.port import PortManager as _PortManager
from vcorelib.io.abc import Serializable as _Serializable

T = _TypeVar("T", bound="AbstractDiGraphNode")
V = _TypeVar("V", bound="AbstractDiGraph")


class AbstractDiGraphNode(_Serializable):
    """A base interface for a directed-graph node."""

    def __init__(
        self, graph: V = None, label: str = None, port: _Port = None, **attrs
    ) -> None:
        """Initialize this graph node."""

        self._graph = graph
        self._label = label
        self.port = port
        self.node_attributes = {**attrs}
        self.ports = _PortManager()

    def add_port(self, label: str, **kwargs) -> _Port:
        """Add a port to this graph node."""
        return self.ports.create(label, **kwargs)

    def allocate_port(self, label: str) -> _Port:
        """Allocate a port on this node."""
        return self.ports.allocate(label)

    def graph(self, kind: _Type[V] = None) -> V:
        """Get the graph that this node belongs to."""
        del kind
        assert self._graph is not None, "Node hasn't joined a graph!"
        return self._graph

    @property
    def label(self) -> str:
        """Get this node's label."""
        assert self._label is not None, "Node hasn't joined a graph!"
        return self._label

    def join_graph(self, graph: V, label: str) -> None:
        """Attempt to join a graph."""

        # Verify the graph assignment.
        assert (
            self._graph is None or self._graph is graph
        ), f"Node '{self}' is already part of a different graph!"

        # Verify the label assignment.
        assert (
            self._label is None or self._label == label
        ), f"Node already has label '{self._label}' (!= '{label}')!"

        self._graph = graph
        self._label = label

    def outgoing(self, graph_kind: _Type[V] = None) -> _Iterator[T]:
        """Iterate over nodes that we have outgoing edges to."""

        graph = self.graph(kind=graph_kind)

        for edge in graph.edges[self.label]:
            yield graph[edge.dst]

    def incoming(self, graph_kind: _Type[V] = None) -> _Iterator[T]:
        """Iterate over nodes that have incoming edges."""

        graph = self.graph(kind=graph_kind)

        for label, edges in graph.edges.items():
            for edge in edges:
                if self.label == edge.dst:
                    yield graph[label]
                    break

    def parallel(self) -> _Set[T]:
        """Iterate over nodes that this instance shares parallel edges with."""

        outgoing: _Set[T] = set(self.outgoing())
        incoming: _Set[T] = set(self.incoming())
        return outgoing.intersection(incoming)


class AbstractDiGraph(
    _UserDict,  # type: ignore
    _Serializable,
):
    """A simple, directed-graph interface."""

    def __init__(
        self,
        name: str,
        initialdata: _Dict[str, T] = None,
        graph_attrs: _AttributeMap = None,
        node_attrs: _AttributeMap = None,
        edge_attrs: _AttributeMap = None,
    ) -> None:
        """Initialize this graph."""

        self.name = name
        super().__init__(initialdata)
        self.graph_attrs = graph_attrs
        self.node_attrs = node_attrs
        self.edge_attrs = edge_attrs

        # Source node -> a set of all destination edges.
        self.edges: _Dict[str, _Set[_GraphEdge]] = {}

    def handle_node(self, label: str, node: T = None) -> T:
        """
        Handle either adding a node as a new vertex or obtaining an existing
        one.
        """

        if node is None:
            assert label in self.data, f"No node '{label}'!"
            node = self.data[label]
        else:
            node = self.add_vertex(label, node)

        return node

    def add_vertex(self, label: str, node: T) -> T:
        """Add a vertext to the graph."""

        assert (
            label not in self.data or self.data[label] is node
        ), f"Duplicate label '{label}'!"

        if label not in self.data:
            # Add a reference for the graph node to the graph itself.
            node.join_graph(self, label)

            # Add the node to the graph.
            self.data[label] = node
            self.edges[label] = set()

        return node

    def add_edge(
        self,
        src: str,
        dst: str,
        src_port: str = None,
        dst_port: str = None,
        strict: bool = False,
        **attrs,
    ) -> None:
        """Add an edge between nodes in the graph."""

        assert src in self.data, f"Source vertex '{src}' not in graph!"
        assert dst in self.data, f"Destination vertex '{dst}' not in graph!"

        # Allocate a source port if a label was provided.
        sport = None
        if src_port:
            sport = self.data[src].allocate_port(src_port)

            # Verify the source port is an output.
            assert (
                sport.is_output
            ), f"Source port '{sport}' ('{src}') is not an output!"

        # Allocate a destination port if a label was provided.
        dport = None
        if dst_port:
            dport = self.data[dst].allocate_port(dst_port)

            # Verify the destination port is an input.
            assert (
                dport.is_input
            ), f"Destination port '{dport}' ('{dst}') is not an input!"

        edge = _GraphEdge(
            src, dst, src_port=sport, dst_port=dport, attrs={**attrs}
        )
        edges = self.edges[src]

        # Prevent adding duplicate edges if desired.
        assert (
            not strict or edge not in edges
        ), f"Edge '{src}' -> '{dst}' already present!"

        edges.add(edge)

    def add_parallel(
        self,
        node1: str,
        node2: str,
        src_port1: str = None,
        dst_port1: str = None,
        src_port2: str = None,
        dst_port2: str = None,
        **kwargs,
    ) -> None:
        """Add an edge in each direction for two nodes."""

        self.add_edge(
            node1, node2, src_port=src_port1, dst_port=dst_port1, **kwargs
        )
        self.add_edge(
            node2, node1, src_port=src_port2, dst_port=dst_port2, **kwargs
        )

    def is_parallel(self, node1: str, node2: str) -> bool:
        """Determine if parallel edges exist between two nodes."""

        success = node1 in self.data and node2 in self.data

        if success:
            success = False

            # Check if we have a destination edge to node1 from node2.
            for edge in self.edges[node1]:
                if edge.dst == node2:
                    success = True
                    break

            if success:
                success = False

                # Check if we have a destination edge to node2 from node1.
                for edge in self.edges[node2]:
                    if edge.dst == node1:
                        success = True
                        break

        return success



"""
A module implementing a file-info cache.
"""

# built-in
from collections.abc import Mapping as _Mapping
from contextlib import contextmanager as _contextmanager
from os import stat_result as _stat_result
from pathlib import Path as _Path
from typing import Callable as _Callable
from typing import Dict as _Dict
from typing import Iterator as _Iterator
from typing import NamedTuple
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import cast as _cast

# internal
from vcorelib.dict.cache import DirectoryCache, FileCache
from vcorelib.io.types import JsonObject as _JsonObject
from vcorelib.paths import Pathlike as _Pathlike
from vcorelib.paths import file_md5_hex as _file_md5_hex
from vcorelib.paths import normalize as _normalize
from vcorelib.paths import stats as _stats


class FileInfo(NamedTuple):
    """A collection of data to describe an identity of a file."""

    path: _Path
    size: int
    md5_hex: str
    modified_ns: int

    def __eq__(self, other: object) -> bool:
        """Determine if two file info's are equivalent."""
        return self.path.samefile(_cast(FileInfo, other).path)

    def __hash__(self) -> int:
        """Get a hash for this file info based on the file name."""
        return hash(str(self.path.resolve()))

    def same(self, other: "FileInfo") -> bool:
        """Check if two file info contents match."""
        return other.size == self.size and other.md5_hex == self.md5_hex

    @staticmethod
    def from_file(path: _Pathlike, stats: _stat_result = None) -> "FileInfo":
        """Create file info from a file."""

        path = _normalize(path).resolve()
        assert path.is_file()
        if stats is None:
            stats = _stats(path)
        assert stats is not None
        md5_hex = _file_md5_hex(path)
        return FileInfo(path, stats.st_size, md5_hex, stats.st_mtime_ns)

    def poll(self) -> _Tuple[bool, _Optional["FileInfo"]]:
        """Determine if this file is in a new state or not."""

        if not self.path.is_file():
            return True, None

        # If the file hasn't been modified, skip re-reading it for a new
        # checksum.
        stats = _stats(self.path)
        if stats is not None and stats.st_mtime_ns == self.modified_ns:
            return False, self

        new_info = FileInfo.from_file(self.path, stats)
        return self.same(new_info), new_info

    def to_json(self, data: _JsonObject = None) -> _JsonObject:
        """Get JSON data for this instance."""
        if data is None:
            data = {}

        path_str = str(self.path.resolve())
        if path_str not in data:
            data[path_str] = {}

        to_update = _cast(_JsonObject, data[path_str])
        to_update["size"] = self.size
        to_update["md5_hex"] = self.md5_hex
        to_update["modified_ns"] = self.modified_ns

        return data

    @staticmethod
    def from_json(data: _JsonObject) -> _Dict[_Path, "FileInfo"]:
        """Create file info from JSON data."""

        result = {}
        for key, info in data.items():
            assert isinstance(info, _Mapping)
            path = _Path(key)
            if path.is_file():
                result[path] = FileInfo(
                    path,
                    _cast(int, info["size"]),
                    _cast(str, info["md5_hex"]),
                    _cast(int, info["modified_ns"]),
                )
        return result


#
# new file (if still present), old file (if it was previously present)
#
FileFreshCallback = _Callable[[_Optional[FileInfo], _Optional[FileInfo]], None]


class FileInfoManager:
    """A class simplifying evaluation of changes to files on disk."""

    def __init__(
        self,
        poll_cb: FileFreshCallback,
        initial: _Dict[_Path, FileInfo] = None,
    ) -> None:
        """Initialize this file-info manager."""

        self.poll = poll_cb

        if initial is None:
            initial = {}
        self.infos = initial

        # If we have initial files, poll them all.
        for path in list(self.infos.keys()):
            self.poll_file(path)

    def poll_directory(self, path: _Pathlike, recurse: bool = True) -> None:
        """Poll an entire directory for changes."""

        norm = _normalize(path).resolve()

        # Don't do anything if this isn't a directory.
        if norm.is_dir():
            for item in norm.iterdir():
                if item.is_file():
                    self.poll_file(item)
                    continue

                if recurse:
                    self.poll_directory(item)

    def poll_file(self, path: _Pathlike) -> None:
        """Check if a file has changed and invoke the callback if so."""

        norm = _normalize(path).resolve()
        old_info = self.infos.get(norm)

        info = None
        if norm not in self.infos:
            info = FileInfo.from_file(norm)
            new = True
        else:
            new, info = self.infos[norm].poll()

        if new:
            self.poll(info, old_info)

        if info is not None:
            self.infos[info.path] = info
        else:
            del self.infos[norm]


@_contextmanager
def file_info_cache(
    cache_path: _Pathlike, poll_cb: FileFreshCallback
) -> _Iterator[FileInfoManager]:
    """Obtain a file-info manager as a cached context."""

    path = _normalize(cache_path)
    cache = DirectoryCache if path.is_dir() else FileCache
    with cache(path).loaded() as data:
        manager = FileInfoManager(poll_cb, FileInfo.from_json(data))
        yield manager

        # Update dictionary data with the current cache contents.
        for info in manager.infos.values():
            info.to_json(data)



"""
A module for analyzing data rates in the time domain.
"""

# built-in
from contextlib import contextmanager
from typing import Iterator as _Iterator

# internal
from vcorelib.math.analysis.average import MovingAverage as _MovingAverage
from vcorelib.math.time import TIMER as _TIMER
from vcorelib.math.time import Timer as _Timer
from vcorelib.math.time import default_time_ns as _default_time_ns


def ns_to_s(nanos: int) -> float:
    """Convert integer nanoseconds to floating-point seconds."""
    return float(nanos / 1e9)


class RateTracker:
    """A class for managing rate information for some data channel."""

    def __init__(self, **kwargs) -> None:
        """Initialize this rate tracker."""

        self.average = _MovingAverage(**kwargs)
        self.prev_time_ns: int = 0

    @property
    def value(self) -> float:
        """An accessor for the underlying value."""
        return self.average.value

    @property
    def min(self) -> float:
        """An accessor for the underlying min."""
        return self.average.min

    @property
    def max(self) -> float:
        """An accessor for the underlying max."""
        return self.average.max

    def __call__(self, time_ns: int = None, value: float = 1.0) -> float:
        """
        Submit new data to the rate tracker. If this function is called with
        default arguments, the returned value will reflect the rate that this
        method is being called in hertz.
        """

        result = 0.0

        # Use a default time if one wasn't provided.
        if time_ns is None:
            time_ns = _default_time_ns()

        # Only start tracking when a second data point is encountered.
        if self.prev_time_ns != 0:
            assert time_ns > self.prev_time_ns

            # Consider 'value' as the amount of change since the last data
            # entry, so divide value by the change in time to get a rate.
            result = self.with_dt(
                ns_to_s(time_ns - self.prev_time_ns), value=value
            )

        self.prev_time_ns = time_ns
        return result

    def with_dt(self, delta_s: float, value: float = 1.0) -> float:
        """Update this rate by directly providing the delta-time value."""
        return self.average(value / delta_s)

    @contextmanager
    def measure(
        self, value: float = 1.0, timer: _Timer = _TIMER
    ) -> _Iterator[None]:
        """Track the time that the caller's context takes."""

        with timer.measure_ns() as token:
            yield
        self.with_dt(ns_to_s(timer.result(token)), value=value)

    def reset(self, initial: float = 0.0) -> None:
        """Reset this rate tracker."""
        self.average.reset(initial=initial)
        self.prev_time_ns = 0



"""
A module for adding newline-related arguments to an argument parser.
"""

# built-in
from argparse import ArgumentParser
from enum import Enum
from os import linesep


class LineEnding(str, Enum):
    """An enumeration for line-ending options."""

    UNIX = "\n"
    DOS = "\r\n"
    PLATFORM = linesep

    def __str__(self) -> str:
        """Get this enum as a string."""
        result = "platform"
        if self is LineEnding.UNIX:
            result = "unix"
        if self is LineEnding.DOS:
            result = "dos"
        return result

    @staticmethod
    def from_arg(opt: str) -> "LineEnding":
        """Convert a string option to an instance of this enum."""
        opt = opt.lower()
        if opt == "unix":
            return LineEnding.UNIX
        if opt == "dos":
            return LineEnding.DOS
        return LineEnding.PLATFORM


def add_newline_arg(parser: ArgumentParser) -> None:
    """Add a line-ending argument to an argument parser."""

    choices = [LineEnding.UNIX, LineEnding.DOS, LineEnding.PLATFORM]
    parser.add_argument(
        "--line-ending",
        type=LineEnding.from_arg,
        default="unix",
        choices=choices,
        help="line-ending option to use by default (default: '%(default)s')",
    )



"""
Utilities for logging information.
"""

# built-in
from contextlib import contextmanager
from logging import INFO as _INFO
from logging import getLogger as _GetLogger
from typing import Iterator as _Iterator

from vcorelib.math.time import LoggerType

# internal
from vcorelib.math.time import TIMER as _TIMER

# Re-export 'LoggerType'.
__all__ = ["LoggerType", "log_time", "LoggerMixin"]


@contextmanager
def log_time(
    log: LoggerType,
    message: str,
    *args,
    level: int = _INFO,
    **kwargs,
) -> _Iterator[None]:
    """
    A simple context manager for conveniently logging time taken for a task.
    """
    with _TIMER.log(log, message, *args, level=level, **kwargs):
        yield


class LoggerMixin:  # pylint: disable=too-few-public-methods
    """A class that provides an inheriting class a logger attribute."""

    logger: LoggerType

    def __init__(self, logger: LoggerType = None) -> None:
        """Initialize this object with logging capabilities."""
        if not hasattr(self, "logger"):
            # Set a logger for this class instance.
            if logger is None:
                logger = _GetLogger(self.__class__.__module__)
            self.logger = logger

    @contextmanager
    def log_time(
        self,
        message: str,
        *args,
        level: int = _INFO,
        **kwargs,
    ) -> _Iterator[None]:
        """A simple wrapper."""
        with log_time(self.logger, message, *args, level=level, **kwargs):
            yield



"""
A module for defining and working with graph edges.
"""

# built-in
from typing import Dict as _Dict
from typing import NamedTuple
from typing import Optional as _Optional
from typing import TextIO as _TextIO

# internal
from vcorelib.graph.port import Port as _Port

AttributeMap = _Dict[str, str]


def write_attributes(stream: _TextIO, data: AttributeMap = None) -> None:
    """A simple attribute writer for Graphviz's DOT language."""

    if data:
        stream.write(" [ ")
        attr_strs = [f'{k}="{v}"' for k, v in data.items()]
        stream.write(", ".join(attr_strs))
        stream.write(" ]")


def write_node_id(label: str, stream: _TextIO, port: _Port = None) -> None:
    """Write a node identifier to a stream."""

    stream.write(label)
    if port:
        stream.write(":")
        stream.write(port.label)


class GraphEdge(NamedTuple):
    """A grouping of attributes describing a directed edge in a graph."""

    src: str
    dst: str
    src_port: _Optional[_Port] = None
    dst_port: _Optional[_Port] = None
    attrs: _Optional[AttributeMap] = None

    def __eq__(self, other) -> bool:
        """Determine if this edge is equivalent to another."""
        return bool(self.src == other.src and self.dst == other.dst)

    def __hash__(self) -> int:
        """Use a string representation as a sane hashing method."""
        return hash(f"{self.src}:{self.dst}")

    def to_stream(self, stream: _TextIO, edgeop: str = "->", **_) -> None:
        """Write this object to a text stream."""

        write_node_id(self.src, stream, port=self.src_port)
        stream.write(f" {edgeop} ")
        write_node_id(self.dst, stream, port=self.dst_port)

        write_attributes(stream, data=self.attrs)
        stream.write(";")



"""
A module for working with application-specific argument parsers.
"""

# built-in
from argparse import ArgumentParser, Namespace
from typing import Callable as _Callable
from typing import Dict as _Dict
from typing import Sequence as _Sequence
from typing import Tuple as _Tuple

CommandFunction = _Callable[[Namespace], int]
CommandRegister = _Callable[[ArgumentParser], CommandFunction]

CommandLoader = _Callable[[], _Sequence[_Tuple[str, str, CommandRegister]]]

CMDS: _Dict[str, CommandFunction] = {}


def app_args(
    command_loader: CommandLoader, commands: _Dict[str, CommandFunction] = None
) -> _Tuple[_Callable[[ArgumentParser], None], CommandFunction]:
    """
    Create a function that can be used to add sub-command processing to an
    argument parser.
    """

    if commands is None:
        commands = CMDS

    def add_app_args(parser: ArgumentParser) -> None:
        """Add application-specific arguments to the command-line parser."""

        subparser = parser.add_subparsers(
            title="commands",
            dest="command",
            required=True,
            help="set of available commands",
        )
        assert commands is not None
        commands.update(
            {
                name: register(subparser.add_parser(name, help=cmd_help))
                for name, cmd_help, register in command_loader()
            }
        )

    def call_command(args: Namespace) -> int:
        """Call the specified command from parsed arguments."""
        assert commands is not None
        return commands[args.command](args)

    return add_app_args, call_command



"""
A module for working with external scripts.
"""

# built-in
import importlib.machinery
import importlib.util
from pathlib import Path as _Path
from typing import Any as _Any
from typing import Set as _Set


def invoke_script(script: _Path, method: str, *args, **kwargs) -> _Any:
    """Invoke a method from an external script."""

    loader = importlib.machinery.SourceFileLoader("script", str(script))
    spec = importlib.util.spec_from_loader("script", loader)
    assert spec is not None
    script_module = importlib.util.module_from_spec(spec)
    loader.exec_module(script_module)

    # Invoke the script method.
    return getattr(script_module, method)(*args, **kwargs)


class ScriptableMixin:
    """
    A mixin for classes to pass themselves to external scripts that may
    update their state.
    """

    def __init__(self) -> None:
        """Initialize this scriptable object."""
        self._invoked: _Set[_Path] = set()

    def script(
        self, script: _Path, method: str, *args, once: bool = True, **kwargs
    ) -> _Any:
        """Invoke a script while passing this object reference."""
        if once:
            assert script not in self._invoked, f"Already invoked '{script}'!"
            self._invoked.add(script)
        return invoke_script(script, method, self, *args, **kwargs)

    def invoked(self, script: _Path) -> bool:
        """Determine if a script has been invoked or not."""
        return script in self._invoked



"""
Common path manipulation utilities.
"""

# built-in
from contextlib import suppress as _suppress
from hashlib import md5 as _md5
from logging import Logger as _Logger
from os import stat_result as _stat_result
from pathlib import Path as _Path
from typing import Iterable as _Iterable
from typing import List as _List
from typing import Optional as _Optional
from typing import Union as _Union

# third-party
from pkg_resources import resource_filename as _resource_filename

# internal
from vcorelib import DEFAULT_ENCODING as _DEFAULT_ENCODING

Pathlike = _Union[_Path, str, None]


def normalize(
    path: Pathlike, *parts: _Union[str, _Path], require: bool = False
) -> _Path:
    """Normalize an input that could be a path into a path."""
    path = (
        _Path("." if path is None else path)
        if not isinstance(path, _Path)
        else path
    )
    path = path.joinpath(*parts)

    if require:
        assert path.exists(), f"Path '{path}' doesn't exist!"

    return path


def stats(path: Pathlike) -> _Optional[_stat_result]:
    """Get stats for a file on disk if it exists."""

    result = None
    path = normalize(path)
    if path.exists():
        result = path.stat()
    return result


def modified_ns(path: Pathlike) -> _Optional[int]:
    """Get the last-modified time from a path if the data can be obtained."""

    result = None
    stat = stats(path)
    if stat is not None:
        result = stat.st_mtime_ns
    return result


def modified_after(path: Pathlike, candidates: _Iterable[Pathlike]) -> bool:
    """
    Check if any candidate paths are more recently modified than the provided
    path. If the path doesn't exists but one or more of the candidates do, this
    method returns True.
    """

    mtime = modified_ns(path)

    # Use zero to compare so that even candidates that exist meet the criteria
    # for being updated after path.
    if mtime is None:
        mtime = 0

    for candidate in candidates:
        curr_mtime = modified_ns(candidate)
        if curr_mtime is not None and curr_mtime > mtime:
            return True

    return False


def get_file_name(path: Pathlike, maxsplit: int = -1) -> str:
    """
    From a path to a file, get the name of the file. Use 'maxsplit' to control
    how many suffixes are considered part of the name or the extension.
    """
    split = normalize(path).name.split(".", maxsplit=maxsplit)
    if len(split) > 1:
        pieces = split[:-1]
    else:
        pieces = [split[0]]
    return ".".join(pieces)


def get_file_ext(path: Pathlike, maxsplit: int = -1) -> str:
    """
    From a path to a file, get the file's extension. Use 'maxsplit' to control
    how many suffixes are considered part of the name or the extension.
    """
    return normalize(path).name.split(".", maxsplit=maxsplit)[-1]


def str_md5_hex(data: str, encoding: str = _DEFAULT_ENCODING) -> str:
    """Get an md5 hex string from string data."""
    return _md5(bytes(data, encoding)).hexdigest()


def file_md5_hex(path: Pathlike) -> str:
    """Get an md5 hex string for a file by path."""
    with normalize(path).open("rb") as stream:
        return _md5(stream.read()).hexdigest()


def _construct_search_path(
    search_paths: _Iterable[Pathlike] = None,
    include_cwd: bool = False,
    relative_to: Pathlike = None,
    package: str = None,
    package_subdir: str = "data",
    logger: _Logger = None,
) -> _List[Pathlike]:
    """Construct a list of paths to search for a resource."""

    to_check: _List[Pathlike] = []

    # Add a package resource to the search path if a package name was provided.
    if package is not None:
        try:
            to_check.append(_resource_filename(package, package_subdir))
        except ModuleNotFoundError:
            if logger is not None:
                logger.warning(
                    "Can't search package '%s', not found.", package
                )

    if search_paths:
        to_check += list(search_paths)

    # Check the working directory if it was specified.
    if include_cwd:
        to_check.append(_Path.cwd())

    # Check if the provided path exists relative to some other path.
    if relative_to is not None:
        relative_to = normalize(relative_to)
        to_check.append(
            relative_to if relative_to.is_dir() else relative_to.parent
        )

    return to_check


def rel(path: Pathlike, base: Pathlike = None) -> _Path:
    """
    Attempt to make 'path' relative to base (which is the current-working
    directory, if not provided).
    """

    path = normalize(path)
    with _suppress(ValueError):
        path = path.relative_to(normalize(base).resolve())
    return path


def find_file(
    path: Pathlike,
    *parts: _Union[str, _Path],
    search_paths: _Iterable[Pathlike] = None,
    include_cwd: bool = False,
    relative_to: Pathlike = None,
    package: str = None,
    package_subdir: str = "data",
    logger: _Logger = None,
) -> _Optional[_Path]:
    """Combines a few simple strategies to locate a file on disk."""

    path = normalize(path, *parts)

    # If path is absolute we can't search for it.
    if path.is_absolute():
        if path.exists():
            return path
        return None

    # Return the first file we find on the search path, if we find one.
    for search in [
        normalize(x)
        for x in _construct_search_path(
            search_paths=search_paths,
            include_cwd=include_cwd,
            relative_to=relative_to,
            package=package,
            package_subdir=package_subdir,
            logger=logger,
        )
    ]:
        if search.is_dir():
            candidate = search.joinpath(path)
            if candidate.exists():
                if logger is not None:
                    logger.debug("Found '%s' at '%s'.", path, search)
                return candidate

            if logger is not None:
                logger.debug("Didn't find '%s' at '%s'.", path, search)

    return None


# An alias for 'find_file' for convenience.
resource = find_file



"""
A module for simplifying limiting the rate of a function call or hot loop.
"""

# built-in
from typing import Callable as _Callable

# internal
from vcorelib.math.analysis.rate import RateTracker
from vcorelib.math.time import default_time_ns as _default_time_ns


class RateLimiter:
    """A class for limiting the rate of runtime work."""

    def __init__(self, period_ns: int, **kwargs) -> None:
        """Initialize this rate-limiter."""

        assert period_ns >= 0
        self.period_ns = period_ns
        self.prev_time_ns: int = 0
        self.rate = RateTracker(**kwargs)

    @property
    def rate_hz(self) -> float:
        """
        Get the underlying rate that this limiter is governing. This is useful
        to determine if the rate limitation is or isn't impacting the rate of
        some task.
        """
        return self.rate.value

    def __call__(
        self, time_ns: int = None, task: _Callable[[], None] = None
    ) -> bool:
        """
        Query the limiter to determine if the current time would allow a
        governed task to run.
        """

        result = False

        # Use a default time if one wasn't provided.
        if time_ns is None:
            time_ns = _default_time_ns()

        if time_ns >= self.prev_time_ns + self.period_ns:
            self.prev_time_ns = time_ns
            result = True

            # Call the task if provided.
            if task is not None:
                task()

            # Update rate tracking.
            self.rate(time_ns=time_ns)

        return result



"""
A simple task management interface.
"""

# built-in
from asyncio import gather
from asyncio import new_event_loop as _new_event_loop
from collections import defaultdict
from typing import Any as _Any
from typing import Callable as _Callable
from typing import Coroutine as _Coroutine
from typing import Dict as _Dict
from typing import Iterable as _Iterable
from typing import List as _List
from typing import Set as _Set
from typing import Tuple as _Tuple
from typing import cast

# internal
from vcorelib.asyncio import run_handle_interrupt as _run_handle_interrupt
from vcorelib.asyncio import shutdown_loop as _shutdown_loop
from vcorelib.dict import merge
from vcorelib.script import ScriptableMixin
from vcorelib.target import TargetMatch
from vcorelib.target.resolver import TargetResolver
from vcorelib.task import Task, TaskFailed

BasicCoroutine = _Callable[[], _Coroutine[_Any, _Any, None]]


class TaskManager(ScriptableMixin):
    """
    A class for managing concurrent execution of tasks and also interfacing
    them via names.
    """

    def __init__(self, resolver: TargetResolver = None) -> None:
        """Initialize this task manager."""

        super().__init__()
        self.tasks: _Dict[str, Task] = {}
        self.dependencies: _Dict[str, _Set[str]] = defaultdict(set)
        self.finalized: bool = False
        if resolver is None:
            resolver = TargetResolver()
        self.resolver = resolver
        self.eloop = _new_event_loop()

    def register(
        self,
        task: Task,
        dependencies: _Iterable[str] = None,
        target: str = None,
    ) -> bool:
        """Register a new task and apply any requested dependencies."""

        new_task = False

        # Register this target so that it will return the provided task if
        # matched.
        if target is None:
            target = task.name

        # Don't take any action if we didn't actually register the task.
        if self.resolver.register(target, task):
            self.tasks[task.name] = task
            new_task = True

        # If we're adding more dependencies, make sure it's for the same task.
        else:
            assert self.tasks[task.name] == task

        if dependencies is None:
            dependencies = set()

        self.dependencies[task.name].update(dependencies)

        # Also add this task's default requirements, if there are any.
        self.dependencies[task.name].update(task.default_requirements)

        self.finalized = False
        return new_task

    def register_to(self, target: str, dependencies: _Iterable[str]) -> bool:
        """Register dependencies to a task by name."""
        return self.register(self.tasks[target], dependencies)

    def finalize(self, **kwargs) -> None:
        """Register task dependencies while the event loop is running."""

        if not self.finalized:
            for task, deps in self.dependencies.items():
                task_obj = self.tasks[task]
                task_obj.dependencies = []
                for dep in deps:
                    # If the dependency points to a task by name, add it.
                    if dep in self.tasks:
                        task_obj.depend_on(
                            self.tasks[dep], eloop=self.eloop, **kwargs
                        )
                        continue

                    # Ensure the dependency can be matched to a task otherwise.
                    resolution = self.resolver.evaluate(dep)
                    assert (
                        resolution
                    ), f"Couldn't match '{dep}' to task '{task}'!"

                    assert resolution.data is not None

                    # Make sure the substitutions from the target resolution
                    # make it to the dependency registration.
                    assert resolution.result.substitutions is not None
                    task_obj.depend_on(
                        cast(Task, resolution.data),
                        eloop=self.eloop,
                        **merge(
                            {**kwargs},
                            resolution.result.substitutions,
                            expect_overwrite=True,
                        ),
                    )

            self.finalized = True

    def evaluate(
        self, tasks: _Iterable[str], **kwargs
    ) -> _Tuple[_List[_Tuple[Task, TargetMatch]], _Set[str]]:
        """
        Iterate over task strings and provide a list of tasks that can be
        executed (plus relevant data from matching the target) as well as
        the set of tasks that can't be resolved.
        """

        # Ensure task dependencies are added.
        self.finalize(**kwargs)

        unresolved: _Set[str] = set()
        task_objs: _List[_Tuple[Task, TargetMatch]] = []
        for resolved in self.resolver.evaluate_all(tasks):
            if isinstance(resolved, str):
                unresolved.add(resolved)
                continue
            task_objs.append((cast(Task, resolved.data), resolved.result))

        return task_objs, unresolved

    def prepare_execute(
        self, tasks: _Iterable[str], **kwargs
    ) -> _Tuple[_Set[str], BasicCoroutine]:
        """
        Gather tasks that can and can't be executed to prepare a function that
        can execute tasks and also return the task strings that wouldn't be
        resolved.
        """

        task_objs, unresolved = self.evaluate(tasks, **kwargs)

        async def executor() -> None:
            """Wait for all of the configured tasks to complete."""

            # Run all tasks together in the event loop.
            await gather(
                *[
                    x[0].dispatch(substitutions=x[1].substitutions, **kwargs)
                    for x in task_objs
                ]
            )

        return unresolved, executor

    def execute(self, tasks: _Iterable[str], **kwargs) -> _Set[str]:
        """
        Execute some set of provided tasks. Return the tasks that don't get
        resolved.
        """

        unresolved, executor = self.prepare_execute(tasks, **kwargs)
        try:
            _run_handle_interrupt(executor(), eloop=self.eloop)

        # Ensure that tasks are cancelled on failure.
        except TaskFailed:
            _shutdown_loop(self.eloop)
            raise

        return unresolved



"""
A module for context managers related to file-system paths.
"""

# built-in
from contextlib import contextmanager, suppress
from os import chdir as _chdir
from os import makedirs as _makedirs
from pathlib import Path as _Path
from tempfile import NamedTemporaryFile as _NamedTemporaryFile
from typing import Iterator as _Iterator
from typing import Union as _Union

# internal
from vcorelib.paths import Pathlike as _Pathlike
from vcorelib.paths import normalize as _normalize


@contextmanager
def in_dir(
    path: _Pathlike, *parts: _Union[str, _Path], makedirs: bool = False
) -> _Iterator[_Path]:
    """Change the current working directory as a context manager."""

    cwd = _Path.cwd()
    try:
        path = _normalize(path, *parts)
        if makedirs:
            _makedirs(path, exist_ok=True)
        _chdir(path)
        yield path
    finally:
        _chdir(cwd)


@contextmanager
def tempfile(*args, **kwargs) -> _Iterator[_Path]:
    """
    Get a valid path to a temporary file and guarantee that its cleaned up
    afterwards.
    """

    with _NamedTemporaryFile(*args, **kwargs) as temp:
        path = _Path(temp.name)
    try:
        yield path
    finally:
        with suppress(FileNotFoundError):
            # Respect the 'delete' argument.
            if kwargs.get("delete", True):
                path.unlink()



"""
A module implementing various data-file decoders.
"""

# built-in
from configparser import ConfigParser, Error, ExtendedInterpolation
from json import load
from json.decoder import JSONDecodeError
from logging import Logger, getLogger
from typing import cast as _cast

# third-party
from ruamel.yaml.parser import ParserError
from ruamel.yaml.scanner import ScannerError
from tomli import TOMLDecodeError, loads

# internal
from vcorelib.dict import consume
from vcorelib.io.types import DataStream as _DataStream
from vcorelib.io.types import JsonObject as _JsonObject
from vcorelib.io.types import LoadResult
from vcorelib.io.types import YAML_INTERFACE as _YAML_INTERFACE
from vcorelib.math.time import TIMER as _TIMER

_LOG = getLogger(__name__)
_INI_INTERPOLATION = ExtendedInterpolation()


def decode_ini(
    data_file: _DataStream,
    logger: Logger = _LOG,
    **kwargs,
) -> LoadResult:
    """Load INI data from a text stream."""

    data = {}
    loaded = True

    with _TIMER.measure_ns() as token:
        # Allow interpolation when reading by default.
        cparser = ConfigParser(
            interpolation=consume(kwargs, "interpolation", _INI_INTERPOLATION),
            **kwargs,
        )
        try:
            cparser.read_file(data_file)
            data = {key: dict(val.items()) for key, val in cparser.items()}
        except Error as exc:
            loaded = False
            logger.error("config-load error: %s", exc)

    return LoadResult(_cast(_JsonObject, data), loaded, _TIMER.result(token))


def decode_json(
    data_file: _DataStream,
    logger: Logger = _LOG,
    **kwargs,
) -> LoadResult:
    """Load JSON data from a text stream."""

    data = {}
    loaded = True

    with _TIMER.measure_ns() as token:
        try:
            data = load(data_file, **kwargs)
            if not data:
                data = {}
        except JSONDecodeError as exc:
            loaded = False
            logger.error("json-load error: %s", exc)

    return LoadResult(data, loaded, _TIMER.result(token))


def decode_yaml(
    data_file: _DataStream,
    logger: Logger = _LOG,
    **kwargs,
) -> LoadResult:
    """Load YAML data from a text stream."""

    data = {}
    loaded = True

    with _TIMER.measure_ns() as token:
        try:
            data = _YAML_INTERFACE.load(data_file, **kwargs)
            if not data:
                data = {}
        except (ScannerError, ParserError) as exc:
            loaded = False
            logger.error("yaml-load error: %s", exc)

    return LoadResult(data, loaded, _TIMER.result(token))


def decode_toml(
    data_file: _DataStream,
    logger: Logger = _LOG,
    **kwargs,
) -> LoadResult:
    """Load TOML data from a text stream."""

    data = {}
    loaded = True

    with _TIMER.measure_ns() as token:
        try:
            data = loads(data_file.read(), **kwargs)
        except TOMLDecodeError as exc:
            loaded = False
            logger.error("toml-load error: %s", exc)

    return LoadResult(data, loaded, _TIMER.result(token))



"""
A module for working with various schema enforcement implementations.
"""

# built-in
from typing import Any as _Any
from typing import Type as _Type

# third-party
from cerberus import Validator as _Validator

# internal
from vcorelib.io.types import JsonObject as _JsonObject
from vcorelib.schemas.base import Schema as _Schema
from vcorelib.schemas.base import SchemaMap as _SchemaMap
from vcorelib.schemas.base import SchemaValidationError


class CerberusSchema(_Schema):
    """An object wrapper for: https://docs.python-cerberus.org/en/stable/."""

    def __init__(self, data: _JsonObject, **kwargs) -> None:
        """Initialize this schema."""
        super().__init__(data)
        self.validator = _Validator(data, **kwargs)

    def __call__(self, data: _Any) -> _Any:
        """Validate input data and return the result."""

        # Normalize the data first.
        data = self.validator.normalized(data)
        if data is not None:
            # Run validation rules.
            if self.validator.validate(data, normalize=False):
                return data

        # Raise error(s).
        raise SchemaValidationError(self.validator.errors)


class CerberusSchemaMap(_SchemaMap):
    """A class for managing multiple schema objects."""

    @classmethod
    def kind(cls) -> _Type[CerberusSchema]:
        """Implement this to determine the concrete schema type."""
        return CerberusSchema



"""
A module defining an interface for dynamic task targets.
"""

from __future__ import annotations

# built-in
import re
from typing import Dict as _Dict
from typing import List as _List
from typing import NamedTuple
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from typing import Union as _union

Substitutions = _Dict[str, _union[str, int]]


class DynamicTargetEvaluator(NamedTuple):
    """
    A regular expression configured to match as many groups as in the provided
    keys. When the pattern matches some data, the names of the keys can become
    associated with the data that was matched inside each group.
    """

    original: str
    pattern: re.Pattern[str]
    keys: _List[str]
    markers: _List[_Tuple[int, int]]

    def compile(self, values: Substitutions) -> str:
        """
        Build a string from this target with values replaced for keys that
        appeared in the original string.
        """

        result = ""
        orig_idx = 0

        for key, marker in zip(self.keys, self.markers):
            result += self.original[orig_idx : marker[0]]
            result += str(values[key])
            orig_idx = marker[1] + 1

        return result + self.original[orig_idx:]


class TargetMatch(NamedTuple):
    """
    An encapsulation of results when attempting to patch a target name to a
    pattern. If a target was matched and had keyword substitutions, the actual
    values used will be set as match data.
    """

    matched: bool
    substitutions: _Optional[Substitutions] = None

    def get(self, data: str) -> str:
        """Get data for keys that matched the target."""
        subs = self.substitutions if self.substitutions is not None else {}
        return str(subs[data])


LITERAL_MATCH = TargetMatch(True)
NO_MATCH = TargetMatch(False)


def escape_regex_special(data: str) -> str:
    """Escape special characters that have meaning in a regular expression."""
    return data.replace(".", "\\.")


class Target:
    """
    An interface for string targets that may encode data substitutions or
    otherwise be matched to only a single, literal string.
    """

    dynamic_start = "{"
    dynamic_end = "}"
    valid = "[a-zA-Z0-9-_.]+"

    def __init__(self, data: str) -> None:
        """Initialize this target."""

        self.data = data
        self.evaluator = self.parse(self.data)
        self.literal = self.evaluator is None

    def __str__(self) -> str:
        """Get this target as a string."""
        return self.data

    def __eq__(self, other: object) -> bool:
        """Check if two targets are equal."""
        return str(self) == str(other)

    def __hash__(self) -> int:
        """Get the hash for this target."""
        return hash(str(self))

    @classmethod
    def is_literal(cls, data: str) -> bool:
        """Determine if a target is guaranteed to be literal or not."""
        return (
            data.count(cls.dynamic_start) == data.count(cls.dynamic_end) == 0
        )

    def compile(self, substitutions: Substitutions = None) -> str:
        """
        Attempt to get a target literal from this target and optional
        substitutions.
        """

        result = self.data
        if self.evaluator is not None:
            assert substitutions is not None, f"Can't compile '{self.data}'!"
            result = self.evaluator.compile(substitutions)
        return result

    @classmethod
    def parse(cls, data: str) -> _Optional[DynamicTargetEvaluator]:
        """
        Obtain a compiled target evaluator if data is a dynamic target, else
        None.
        """

        # The short-circuit case where this is not a dynamic target.
        if cls.is_literal(data):
            return None

        # A preliminary syntax check.
        open_len = data.count(cls.dynamic_start)
        assert open_len == data.count(cls.dynamic_end)

        pattern = "^"
        keys = []
        markers: _List[_Tuple[int, int]] = []
        live = data
        abs_idx = 0
        for _ in range(open_len):
            start = live.index(cls.dynamic_start)
            end = live.index(cls.dynamic_end)

            # Store the absolute index into the string that the control
            # characters appeared.
            markers.append((abs_idx + start, abs_idx + end))

            pattern += escape_regex_special(live[:start])
            pattern += f"({cls.valid})"

            keys.append(live[start + 1 : end])
            live = live[end + 1 :]
            abs_idx += end + 1

        pattern += escape_regex_special(live) + "$"

        assert len(keys) == open_len
        return DynamicTargetEvaluator(data, re.compile(pattern), keys, markers)

    def evaluate(self, data: str) -> TargetMatch:
        """Attempt to match this target with some string data."""

        if self.literal:
            return TargetMatch(self.data == data)

        assert self.evaluator is not None
        result = self.evaluator.pattern.fullmatch(data)
        if result is None:
            return NO_MATCH

        return TargetMatch(
            True,
            {
                key: result.group(1 + idx)
                for idx, key in enumerate(self.evaluator.keys)
            },
        )



"""
A module for implementing schema-validated classes.
"""

# internal
from vcorelib.schemas.base import Schema as _Schema
from vcorelib.schemas.base import SchemaMap as _SchemaMap


class SchemaMixin:
    """
    A class that allows inheriting classes to validate an attribute with a
    schema.
    """

    schema: _Schema

    def __init__(self, schemas: _SchemaMap, valid_attr: str = "data") -> None:
        """Initialize this object instance by performing schema validation."""

        # Don't double initialize.
        if not hasattr(self, "schema"):
            # Allow the name of the schema to be overridden if necessary.
            schema = self.schema_name()
            assert schema in schemas, f"No schema for '{schema}'!"
            self.schema = schemas[schema]

            # Perform validation.
            self.validate(valid_attr=valid_attr)

    @classmethod
    def schema_name(cls) -> str:
        """A default name for this class's schema."""
        return cls.__name__

    def validate(self, valid_attr: str = "data") -> None:
        """Validate an instance attribute based on this instance's schema."""

        # Ensure that this object has the specified attribute.
        assert hasattr(
            self, valid_attr
        ), f"Object doesn't have attribute '{valid_attr}' to validate!"

        # Validate object data. Re-assign the result.
        setattr(self, valid_attr, self.schema(getattr(self, valid_attr)))



"""
A module for cache implementations - conforming to package-wide, data-structure
constraints and assumptions.
"""

# built-in
from collections import UserDict
from logging import DEBUG as _DEBUG
from logging import Logger
from pathlib import Path
from shutil import rmtree
from typing import MutableMapping as _MutableMapping

# third-party
from vcorelib.dict import merge
from vcorelib.io import ARBITER as _ARBITER
from vcorelib.io import DataArbiter

# internal
from vcorelib.io.archive import extractall, make_archive
from vcorelib.io.types import DEFAULT_ARCHIVE_EXT as _DEFAULT_ARCHIVE_EXT
from vcorelib.io.types import DEFAULT_DATA_EXT as _DEFAULT_DATA_EXT
from vcorelib.io.types import FileExtension
from vcorelib.io.types import JsonObject as _JsonObject
from vcorelib.io.types import JsonValue as _JsonValue
from vcorelib.math.time import TIMER as _TIMER
from vcorelib.math.time import byte_count_str, nano_str
from vcorelib.paths import Pathlike as _Pathlike


class FlatDirectoryCache(
    UserDict,  # type: ignore
    _MutableMapping[str, _JsonValue],
):
    """
    A class implementing a dictionary that can be saved and loaded from disk,
    with a specified encoding scheme.
    """

    def __init__(
        self,
        root: Path,
        initialdata: _JsonObject = None,
        archive_encoding: str = _DEFAULT_ARCHIVE_EXT,
        data_encoding: str = _DEFAULT_DATA_EXT,
        arbiter: DataArbiter = _ARBITER,
        **load_kwargs,
    ) -> None:
        """Initialize this data cache."""

        super().__init__(initialdata)
        self.root = root
        self.archive_encoding = archive_encoding
        self.data_encoding = data_encoding
        self.arbiter = arbiter
        self.load_time_ns: int = -1
        self.save_time_ns: int = -1

        # A derived class must add logic to set this.
        self.changed: bool = False

        merge(self.data, self.load(self.root, **load_kwargs))

    def load_directory(
        self,
        path: _Pathlike,
        data: _JsonObject,
        **kwargs,
    ) -> int:
        """Load a directory and update data, return time taken to load."""

        load = self.arbiter.decode_directory(
            path, require_success=True, **kwargs
        )
        data.update(load.data)
        return load.time_ns

    def load(
        self,
        path: Path = None,
        logger: Logger = None,
        level: int = _DEBUG,
        **kwargs,
    ) -> _JsonObject:
        """Load data from disk."""

        if path is None:
            path = self.root

        loaded = False
        result: _JsonObject = {}
        if path.is_dir():
            self.load_time_ns = self.load_directory(path, result, **kwargs)
            loaded = True

        # See if we can locate an archive for this path, that we can extract
        # and then load.
        else:
            archive = FileExtension.has_archive(path)
            if archive is not None:
                success, time_ns = extractall(archive, path.parent)
                if success:
                    if logger is not None:
                        logger.log(
                            level,
                            "Extracted archive '%s' in %ss.",
                            archive,
                            nano_str(time_ns, True),
                        )
                    return self.load(path, logger, level, **kwargs)

        if loaded and logger is not None:
            logger.log(
                level,
                "Cache loaded in %ss.",
                nano_str(self.load_time_ns, True),
            )
        return result

    def save_directory(self, path: Path, **kwargs) -> int:
        """Write data in this cache to a directory."""

        path.mkdir(parents=True, exist_ok=True)
        return self.arbiter.encode_directory(
            path, self.data, self.data_encoding, **kwargs
        )[1]

    def save(
        self,
        path: Path = None,
        logger: Logger = None,
        level: int = _DEBUG,
        archive: bool = False,
        **kwargs,
    ) -> None:
        """Save data to disk."""

        if path is None:
            path = self.root

        if self.changed:
            self.save_time_ns = self.save_directory(path, **kwargs)

        # Create an archive for this cache if requested.
        if archive:
            result = make_archive(path, self.archive_encoding)
            assert result[0] is not None, "Tried to make archive but couldn't!"
            if logger is not None:
                logger.log(
                    level,
                    "Cache archived to '%s' (%s) in %ss.",
                    result[0],
                    byte_count_str(result[0].stat().st_size),
                    nano_str(result[1], True),
                )

        if self.changed and logger is not None:
            logger.log(
                level,
                "Cache written in %ss.",
                nano_str(self.save_time_ns, True),
            )
        self.changed = False

    def clean(
        self,
        path: Path = None,
        logger: Logger = None,
        level: int = _DEBUG,
    ) -> None:
        """Remove cached data from disk."""

        if path is None:
            path = self.root

        with _TIMER.measure_ns() as token:
            # Remove any archives.
            for candidate in FileExtension.archive_candidates(path, True):
                candidate.unlink()

            # Remove the data directory.
            rmtree(path, ignore_errors=True)

        time_ns = _TIMER.result(token)
        if logger is not None:
            logger.log(level, "Cache cleaned in %ss.", nano_str(time_ns, True))
